# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

from drgn.helpers.linux.panic import _panic_message, tainted
from tests import TestCase
from tests.linux_kernel import LinuxKernelTestCase, skip_unless_have_test_kmod


class TestPanicMessage(TestCase):
    def test_basic(self):
        self.assertEqual(
            _panic_message(
                b"""\
Freeing unused kernel image (text/rodata gap) memory: 1748K
Freeing unused kernel image (rodata/data gap) memory: 1300K
Run /tmp/drgn-vmtest-bwe4pfi5/init as init process
sysrq: Trigger a crash
Kernel panic - not syncing: sysrq triggered crash
CPU: 6 UID: 0 PID: 127 Comm: selfdestruct Kdump: loaded Not tainted 6.16.0-rc5-vmtest35.1default #1 PREEMPT(none)
Hardware name: QEMU Standard PC (i440FX + PIIX, 1996), BIOS 1.17.0-4.fc42 04/01/2014
"""
            ),
            b"Kernel panic - not syncing: sysrq triggered crash",
        )

    def test_preference(self):
        # We should prefer BUG over Oops.
        self.assertEqual(
            _panic_message(
                b"""\
Run /tmp/drgn-vmtest-50jcxww4/init as init process
drgn_test: loading out-of-tree module taints kernel.
stackdepot: allocating hash table of 131072 entries via kvcalloc
BUG: kernel NULL pointer dereference, address: 000000000000071c
#PF: supervisor write access in kernel mode
#PF: error_code(0x0002) - not-present page
PGD 0 P4D 0
Oops: Oops: 0002 [#1] SMP NOPTI
"""
            ),
            b"BUG: kernel NULL pointer dereference, address: 000000000000071c",
        )


class TestTainted(LinuxKernelTestCase):
    @skip_unless_have_test_kmod
    def test_tainted(self):
        self.assertIn("O", tainted(self.prog))



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import itertools
from typing import Generic, Iterator, List, Mapping, Sequence, TypeVar, Union

from drgndoc.parse import (
    Class,
    DocumentedNode,
    Function,
    Import,
    ImportFrom,
    Module,
    Node,
    Variable,
)

NodeT_co = TypeVar("NodeT_co", bound=Node, covariant=True)


class BoundNode(Generic[NodeT_co]):
    def __init__(self, name: str, node: NodeT_co) -> None:
        self.name = name
        self.node = node


class ResolvedNode(Generic[NodeT_co]):
    def __init__(
        self,
        modules: Sequence[BoundNode[Module]],
        classes: Sequence[BoundNode[Class]],
        name: str,
        node: NodeT_co,
    ) -> None:
        self.modules = modules
        self.classes = classes
        self.name = name
        self.node = node

    def qualified_name(self) -> str:
        return ".".join(
            itertools.chain(
                (module.name for module in self.modules),
                (class_.name for class_ in self.classes),
                (self.name,),
            )
        )

    def attrs(self) -> Iterator["ResolvedNode[Node]"]:
        if isinstance(self.node, Module):
            modules = list(self.modules)
            modules.append(BoundNode(self.name, self.node))
            for attr, node in self.node.attrs.items():
                yield ResolvedNode(modules, self.classes, attr, node)
        elif isinstance(self.node, Class):
            classes = list(self.classes)
            classes.append(BoundNode(self.name, self.node))
            for attr, node in self.node.attrs.items():
                yield ResolvedNode(self.modules, classes, attr, node)

    def attr(self, attr: str) -> "ResolvedNode[Node]":
        if isinstance(self.node, Module):
            modules = list(self.modules)
            modules.append(BoundNode(self.name, self.node))
            return ResolvedNode(modules, self.classes, attr, self.node.attrs[attr])
        elif isinstance(self.node, Class):
            classes = list(self.classes)
            classes.append(BoundNode(self.name, self.node))
            return ResolvedNode(self.modules, classes, attr, self.node.attrs[attr])
        else:
            raise KeyError(attr)


UnresolvedName = str


class Namespace:
    def __init__(self, modules: Mapping[str, Module]) -> None:
        self.modules = modules

    # NB: this modifies the passed lists.
    def _resolve_name(
        self,
        modules: List[BoundNode[Module]],
        classes: List[BoundNode[Class]],
        name_components: List[str],
    ) -> Union[ResolvedNode[DocumentedNode], UnresolvedName]:
        name_components.reverse()
        while name_components:
            attrs: Mapping[str, Node]
            if classes:
                attrs = classes[-1].node.attrs
            elif modules:
                attrs = modules[-1].node.attrs
            else:
                attrs = self.modules
            name = name_components.pop()
            try:
                node = attrs[name]
            except KeyError:
                break

            if isinstance(node, (Import, ImportFrom)):
                classes.clear()
                if isinstance(node, Import):
                    modules.clear()
                elif isinstance(node, ImportFrom):
                    if node.level >= len(modules):
                        # Relative import beyond top-level package. Bail.
                        break
                    # Absolute import is level 0, which clears the whole list.
                    del modules[-node.level :]
                    name_components.append(node.name)
                if node.module is not None:
                    name_components.extend(reversed(node.module.split(".")))
            elif name_components:
                if isinstance(node, Module):
                    assert not classes
                    modules.append(BoundNode(name, node))
                elif isinstance(node, Class):
                    classes.append(BoundNode(name, node))
                else:
                    break
        else:
            assert isinstance(node, (Module, Class, Function, Variable))
            return ResolvedNode(modules, classes, name, node)
        return ".".join(
            itertools.chain(
                (module.name for module in modules),
                (class_.name for class_ in classes),
                (name,),
                reversed(name_components),
            )
        )

    def resolve_global_name(
        self, name: str
    ) -> Union[ResolvedNode[DocumentedNode], UnresolvedName]:
        return self._resolve_name([], [], name.split("."))

    def resolve_name_in_scope(
        self,
        modules: Sequence[BoundNode[Module]],
        classes: Sequence[BoundNode[Class]],
        name: str,
    ) -> Union[ResolvedNode[DocumentedNode], UnresolvedName]:
        name_components = name.split(".")
        attr = name_components[0]
        if classes and attr in classes[-1].node.attrs:
            classes = list(classes)
        elif modules and attr in modules[-1].node.attrs:
            classes = []
        else:
            return name
        modules = list(modules)
        return self._resolve_name(modules, classes, name_components)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

REGISTERS = [
    DrgnRegister("rax"),
    DrgnRegister("rdx"),
    DrgnRegister("rcx"),
    DrgnRegister("rbx"),
    DrgnRegister("rsi"),
    DrgnRegister("rdi"),
    DrgnRegister("rbp"),
    DrgnRegister("rsp"),
    DrgnRegister("r8"),
    DrgnRegister("r9"),
    DrgnRegister("r10"),
    DrgnRegister("r11"),
    DrgnRegister("r12"),
    DrgnRegister("r13"),
    DrgnRegister("r14"),
    DrgnRegister("r15"),
    DrgnRegister("rip"),
    DrgnRegister("rflags"),
    DrgnRegister("es"),
    DrgnRegister("cs"),
    DrgnRegister("ss"),
    DrgnRegister("ds"),
    DrgnRegister("fs"),
    DrgnRegister("gs"),
    DrgnRegister("fs.base"),
    DrgnRegister("gs.base"),
]

REGISTER_LAYOUT = [
    # The psABI calls this the return address (RA) register.
    DrgnRegisterLayout("rip", size=8, dwarf_number=16),
    DrgnRegisterLayout("rsp", size=8, dwarf_number=7),
    # General-purpose registers in the same order as struct pt_regs (which has
    # callee-saved registers first).
    DrgnRegisterLayout("r15", size=8, dwarf_number=15),
    DrgnRegisterLayout("r14", size=8, dwarf_number=14),
    DrgnRegisterLayout("r13", size=8, dwarf_number=13),
    DrgnRegisterLayout("r12", size=8, dwarf_number=12),
    DrgnRegisterLayout("rbp", size=8, dwarf_number=6),
    DrgnRegisterLayout("rbx", size=8, dwarf_number=3),
    DrgnRegisterLayout("r11", size=8, dwarf_number=11),
    DrgnRegisterLayout("r10", size=8, dwarf_number=10),
    DrgnRegisterLayout("r9", size=8, dwarf_number=9),
    DrgnRegisterLayout("r8", size=8, dwarf_number=8),
    DrgnRegisterLayout("rax", size=8, dwarf_number=0),
    DrgnRegisterLayout("rcx", size=8, dwarf_number=2),
    DrgnRegisterLayout("rdx", size=8, dwarf_number=1),
    DrgnRegisterLayout("rsi", size=8, dwarf_number=4),
    DrgnRegisterLayout("rdi", size=8, dwarf_number=5),
    # Segment and flags registers in the same order as struct user_regs_struct
    # (except for orig_rax, rip, and rsp).
    # orig_rax and rip are here in struct user_regs_struct.
    DrgnRegisterLayout("cs", size=8, dwarf_number=51),
    DrgnRegisterLayout("rflags", size=8, dwarf_number=49),
    # rsp is here in struct user_regs_struct.
    DrgnRegisterLayout("ss", size=8, dwarf_number=52),
    DrgnRegisterLayout("fs_base", size=8, dwarf_number=58),
    DrgnRegisterLayout("gs_base", size=8, dwarf_number=59),
    DrgnRegisterLayout("ds", size=8, dwarf_number=53),
    DrgnRegisterLayout("es", size=8, dwarf_number=50),
    DrgnRegisterLayout("fs", size=8, dwarf_number=54),
    DrgnRegisterLayout("gs", size=8, dwarf_number=55),
]

STACK_POINTER_REGISTER = "rsp"



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later
import itertools

from drgn import Architecture, Platform, PlatformFlags
from tests import TestCase


class TestPlatform(TestCase):
    def test_default_flags(self):
        Platform(Architecture.X86_64)
        self.assertRaises(ValueError, Platform, Architecture.UNKNOWN)

    def test_registers(self):
        self.assertIn(
            "rax",
            itertools.chain.from_iterable(
                reg.names for reg in Platform(Architecture.X86_64).registers
            ),
        )
        self.assertEqual(Platform(Architecture.UNKNOWN, PlatformFlags(0)).registers, ())



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import os
import struct
from typing import List, NamedTuple, Optional, Sequence, Tuple, Union
import zlib

from _drgn_util.elf import ET, PT, SHF, SHN, SHT, STB, STT, STV


class ElfSection:
    def __init__(
        self,
        *,
        data: bytes = b"",
        name: Optional[str] = None,
        sh_type: Optional[SHT] = None,
        p_type: Optional[PT] = None,
        vaddr: int = 0,
        paddr: int = 0,
        memsz: Optional[int] = None,
        p_align: int = 0,
        sh_link: int = 0,
        sh_info: int = 0,
        sh_entsize: int = 0,
        sh_flags: SHF = SHF(0),
    ):
        self.data = data
        self.name = name
        self.sh_type = sh_type
        self.sh_flags = sh_flags
        self.p_type = p_type
        self.vaddr = vaddr
        self.paddr = paddr
        self.memsz = memsz
        self.p_align = p_align
        self.sh_link = sh_link
        self.sh_info = sh_info
        self.sh_entsize = sh_entsize

        assert (self.name is not None) or (self.p_type is not None)
        assert (self.name is None) == (self.sh_type is None)
        assert self.p_type is None or not (sh_flags & SHF.COMPRESSED)


class ElfSymbol(NamedTuple):
    name: str
    value: int
    size: int
    type: STT
    binding: STB
    shindex: Optional[int] = None
    visibility: STV = STV.DEFAULT

    def st_info(self) -> int:
        return (self.binding << 4) + (self.type & 0xF)


def _create_symtab(
    sections: List[ElfSection],
    symbols: Sequence[ElfSymbol],
    *,
    dynamic: bool = False,
    little_endian: bool,
    bits: int,
):
    symtab_name = ".dynsym" if dynamic else ".symtab"
    strtab_name = ".dynstr" if dynamic else ".strtab"
    assert not any(section.name in (symtab_name, strtab_name) for section in sections)

    # An empty symbol name is a placeholder for the implicit 0-index entry in
    # the symbol table. It's used to create a valid, but empty symbol table.
    if symbols and symbols[0].name == "":
        symbols = symbols[1:]

    endian = "<" if little_endian else ">"
    if bits == 64:
        symbol_struct = struct.Struct(endian + "IBBHQQ")

        def symbol_fields(sym: ElfSymbol):
            return (
                sym.st_info(),
                sym.visibility,
                SHN.UNDEF if sym.shindex is None else sym.shindex,
                sym.value,
                sym.size,
            )

    else:
        symbol_struct = struct.Struct(endian + "IIIBBH")

        def symbol_fields(sym: ElfSymbol):
            return (
                sym.value,
                sym.size,
                sym.st_info(),
                sym.visibility,
                SHN.UNDEF if sym.shindex is None else sym.shindex,
            )

    symtab_data = bytearray((len(symbols) + 1) * symbol_struct.size)
    strtab_data = bytearray(1)
    sh_info = 1
    for i, sym in enumerate(symbols, 1):
        symbol_struct.pack_into(
            symtab_data, i * symbol_struct.size, len(strtab_data), *symbol_fields(sym)
        )
        strtab_data.extend(sym.name.encode())
        strtab_data.append(0)
        if sym.binding == STB.LOCAL:
            assert sh_info == i, "local symbol after non-local symbol"
            sh_info = i + 1

    sections.append(
        ElfSection(
            name=symtab_name,
            sh_type=SHT.DYNSYM if dynamic else SHT.SYMTAB,
            data=symtab_data,
            sh_link=sum((1 for section in sections if section.name is not None), 2),
            sh_info=sh_info,
            sh_entsize=symbol_struct.size,
        )
    )
    sections.append(ElfSection(name=strtab_name, sh_type=SHT.STRTAB, data=strtab_data))


def create_elf_file(
    type: ET,
    sections: Sequence[ElfSection] = (),
    symbols: Sequence[ElfSymbol] = (),
    *,
    dynamic_symbols: Sequence[ElfSymbol] = (),
    build_id: Optional[bytes] = None,
    gnu_debuglink: Optional[
        Tuple[Union[str, bytes, "os.PathLike[str]", "os.PathLike[bytes]"], int]
    ] = None,
    gnu_debugaltlink: Optional[
        Tuple[Union[str, bytes, "os.PathLike[str]", "os.PathLike[bytes]"], bytes]
    ] = None,
    little_endian: bool = True,
    bits: int = 64,
):
    endian = "<" if little_endian else ">"
    if bits == 64:
        ehdr_struct = struct.Struct(endian + "16BHHIQQQIHHHHHH")
        shdr_struct = struct.Struct(endian + "IIQQQQIIQQ")
        phdr_struct = struct.Struct(endian + "IIQQQQQQ")
        chdr_struct = struct.Struct(endian + "IIQQ")
        e_machine = 62 if little_endian else 43  # EM_X86_64 or EM_SPARCV9
    else:
        assert bits == 32
        ehdr_struct = struct.Struct(endian + "16BHHIIIIIHHHHHH")
        shdr_struct = struct.Struct(endian + "10I")
        phdr_struct = struct.Struct(endian + "8I")
        chdr_struct = struct.Struct(endian + "III")
        e_machine = 3 if little_endian else 8  # EM_386 or EM_MIPS
    nhdr_struct = struct.Struct(endian + "3I")

    sections = list(sections)
    if dynamic_symbols:
        _create_symtab(
            sections,
            dynamic_symbols,
            dynamic=True,
            little_endian=little_endian,
            bits=bits,
        )
    if symbols:
        _create_symtab(sections, symbols, little_endian=little_endian, bits=bits)
    if build_id is not None:
        build_id_note = (
            nhdr_struct.pack(
                4,  # n_namesz,
                len(build_id),  # n_namesz,
                3,  # n_type = NT_GNU_BUILD_ID
            )
            + b"GNU\0"
            + build_id
            + bytes(-len(build_id) % 4)
        )
        sections.append(
            ElfSection(name=".note.gnu.build-id", sh_type=SHT.NOTE, data=build_id_note)
        )

    if gnu_debuglink is not None:
        gnu_debuglink_path, gnu_debuglink_crc = gnu_debuglink
        gnu_debuglink_path = os.fsencode(gnu_debuglink_path)
        sections.append(
            ElfSection(
                name=".gnu_debuglink",
                sh_type=SHT.PROGBITS,
                data=gnu_debuglink_path
                + bytes(4 - len(gnu_debuglink_path) % 4)
                + gnu_debuglink_crc.to_bytes(4, "little"),
            )
        )

    if gnu_debugaltlink is not None:
        gnu_debugaltlink_path, gnu_debugaltlink_build_id = gnu_debugaltlink
        sections.append(
            ElfSection(
                name=".gnu_debugaltlink",
                sh_type=SHT.PROGBITS,
                data=os.fsencode(gnu_debugaltlink_path)
                + b"\0"
                + gnu_debugaltlink_build_id,
            )
        )

    shnum = 0
    phnum = 0
    shstrtab = bytearray(1)
    for section in sections:
        if section.name is not None:
            shstrtab.extend(section.name.encode())
            shstrtab.append(0)
            shnum += 1
        if section.p_type is not None:
            phnum += 1
    if shnum > 0:
        shnum += 2  # One for the SHT_NULL section, one for .shstrtab.
        shstrtab.extend(b".shstrtab\0")
        sections.append(ElfSection(name=".shstrtab", sh_type=SHT.STRTAB, data=shstrtab))

    shdr_offset = ehdr_struct.size
    phdr_offset = shdr_offset + shdr_struct.size * shnum
    headers_size = phdr_offset + phdr_struct.size * phnum
    buf = bytearray(headers_size)
    ehdr_struct.pack_into(
        buf,
        0,
        0x7F,  # ELFMAG0
        ord("E"),  # ELFMAG1
        ord("L"),  # ELFMAG2
        ord("F"),  # ELFMAG3
        2 if bits == 64 else 1,  # EI_CLASS = ELFCLASS64 or ELFCLASS32
        1 if little_endian else 2,  # EI_DATA = ELFDATA2LSB or ELFDATA2MSB
        1,  # EI_VERSION = EV_CURRENT
        0,  # EI_OSABI = ELFOSABI_NONE
        0,  # EI_ABIVERSION
        0,
        0,
        0,
        0,
        0,
        0,
        0,  # EI_PAD
        type,  # e_type
        e_machine,
        1,  # e_version = EV_CURRENT
        0,  # e_entry
        phdr_offset if phnum else 0,  # e_phoff
        shdr_offset if shnum else 0,  # e_shoff
        0,  # e_flags
        ehdr_struct.size,  # e_ehsize
        phdr_struct.size,  # e_phentsize
        phnum,  # e_phnum
        shdr_struct.size,  # e_shentsize
        shnum,  # e_shnum
        shnum - 1 if shnum else 0,  # e_shstrndx
    )

    shdr_offset += shdr_struct.size
    for section in sections:
        ch_addralign = 1 if section.p_type is None else bits // 8
        memsz = len(section.data) if section.memsz is None else section.memsz
        if section.sh_flags & SHF.COMPRESSED:
            sh_addralign = bits // 8
            compressed_data = zlib.compress(section.data)
            sh_size = chdr_struct.size + len(compressed_data)
        else:
            sh_addralign = ch_addralign
            sh_size = memsz
        if section.p_align:
            padding = section.vaddr % section.p_align - len(buf) % section.p_align
            buf.extend(bytes(padding))
        if section.name is not None:
            shdr_struct.pack_into(
                buf,
                shdr_offset,
                shstrtab.index(section.name.encode()),  # sh_name
                section.sh_type,  # sh_type
                section.sh_flags,  # sh_flags
                section.vaddr,  # sh_addr
                len(buf),  # sh_offset
                sh_size,  # sh_size
                section.sh_link,  # sh_link
                section.sh_info,  # sh_info
                sh_addralign,  # sh_addralign
                section.sh_entsize,  # sh_entsize
            )
            shdr_offset += shdr_struct.size
        if section.p_type is not None:
            flags = 7  # PF_R | PF_W | PF_X
            if bits == 64:
                phdr_struct.pack_into(
                    buf,
                    phdr_offset,
                    section.p_type,  # p_type
                    flags,  # p_flags
                    len(buf),  # p_offset
                    section.vaddr,  # p_vaddr
                    section.paddr,  # p_paddr
                    len(section.data),  # p_filesz
                    memsz,  # p_memsz
                    section.p_align,  # p_align
                )
            else:
                phdr_struct.pack_into(
                    buf,
                    phdr_offset,
                    section.p_type,  # p_type
                    len(buf),  # p_offset
                    section.vaddr,  # p_vaddr
                    section.paddr,  # p_paddr
                    len(section.data),  # p_filesz
                    memsz,  # p_memsz
                    flags,  # p_flags
                    section.p_align,  # p_align
                )
            phdr_offset += phdr_struct.size
        if section.sh_flags & SHF.COMPRESSED:
            ELFCOMPRESS_ZLIB = 1
            if bits == 64:
                buf.extend(
                    chdr_struct.pack(
                        ELFCOMPRESS_ZLIB,  # ch_type
                        0,  # ch_reserved
                        memsz,  # ch_size
                        ch_addralign,  # ch_addralign
                    )
                )
            else:
                buf.extend(
                    chdr_struct.pack(
                        ELFCOMPRESS_ZLIB,  # ch_type
                        memsz,  # ch_size
                        ch_addralign,  # ch_addralign
                    )
                )
            buf.extend(compressed_data)
        else:
            buf.extend(section.data)

    return buf



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import argparse
import re
import subprocess

# Macros missing from glibc 2.17, which is the oldest version we test against
# (when we build manylinux2014 wheels).
MACROS = (
    # Added in glibc 2.18.
    "NT_FILE",
    # Added in glibc 2.24.
    "EM_RISCV",
    # Added in glibc 2.27.
    # We don't need all of the RISC-V relocation types, but it doesn't hurt.
    "R_RISCV_NONE",
    "R_RISCV_32",
    "R_RISCV_64",
    "R_RISCV_RELATIVE",
    "R_RISCV_COPY",
    "R_RISCV_JUMP_SLOT",
    "R_RISCV_TLS_DTPMOD32",
    "R_RISCV_TLS_DTPMOD64",
    "R_RISCV_TLS_DTPREL32",
    "R_RISCV_TLS_DTPREL64",
    "R_RISCV_TLS_TPREL32",
    "R_RISCV_TLS_TPREL64",
    # Added in glibc 2.28.
    "R_RISCV_BRANCH",
    "R_RISCV_JAL",
    "R_RISCV_CALL",
    "R_RISCV_CALL_PLT",
    "R_RISCV_GOT_HI20",
    "R_RISCV_TLS_GOT_HI20",
    "R_RISCV_TLS_GD_HI20",
    "R_RISCV_PCREL_HI20",
    "R_RISCV_PCREL_LO12_I",
    "R_RISCV_PCREL_LO12_S",
    "R_RISCV_HI20",
    "R_RISCV_LO12_I",
    "R_RISCV_LO12_S",
    "R_RISCV_TPREL_HI20",
    "R_RISCV_TPREL_LO12_I",
    "R_RISCV_TPREL_LO12_S",
    "R_RISCV_TPREL_ADD",
    "R_RISCV_ADD8",
    "R_RISCV_ADD16",
    "R_RISCV_ADD32",
    "R_RISCV_ADD64",
    "R_RISCV_SUB8",
    "R_RISCV_SUB16",
    "R_RISCV_SUB32",
    "R_RISCV_SUB64",
    "R_RISCV_GNU_VTINHERIT",
    "R_RISCV_GNU_VTENTRY",
    "R_RISCV_ALIGN",
    "R_RISCV_RVC_BRANCH",
    "R_RISCV_RVC_JUMP",
    "R_RISCV_RVC_LUI",
    "R_RISCV_GPREL_I",
    "R_RISCV_GPREL_S",
    "R_RISCV_TPREL_I",
    "R_RISCV_TPREL_S",
    "R_RISCV_RELAX",
    "R_RISCV_SUB6",
    "R_RISCV_SET6",
    "R_RISCV_SET8",
    "R_RISCV_SET16",
    "R_RISCV_SET32",
    "R_RISCV_32_PCREL",
    # Added in glibc 2.30.
    "NT_ARM_PAC_MASK",
)


def main() -> None:
    argparse.ArgumentParser(
        description="Generate definitions missing from older versions of glibc for libdrgn/include/elf.h from elf.h"
    ).parse_args()

    contents = subprocess.check_output(
        ["gcc", "-dD", "-E", "-"],
        input="#include <elf.h>\n",
        universal_newlines=True,
    )

    macros = {
        match.group(1): match.group(2)
        for match in re.finditer(
            r"^\s*#\s*define\s+(" + "|".join(MACROS) + r")\s+(.*)",
            contents,
            re.MULTILINE,
        )
    }

    print("// Generated by scripts/gen_elf_compat.py.")
    for macro in MACROS:
        print("#ifndef", macro)
        print("#define", macro, macros[macro])
        print("#endif")


if __name__ == "__main__":
    main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

from multiprocessing import Barrier, Process
import os

from drgn.helpers.linux.pid import find_task
from tests.linux_kernel import LinuxKernelTestCase


class TestThreads(LinuxKernelTestCase):
    def test_threads(self):
        NUM_PROCS = 12
        barrier = Barrier(NUM_PROCS + 1)

        try:
            procs = [Process(target=barrier.wait) for _ in range(NUM_PROCS)]
            for proc in procs:
                proc.start()
            pids = {thread.tid for thread in self.prog.threads()}
            for proc in procs:
                self.assertIn(proc.pid, pids)
            self.assertIn(os.getpid(), pids)
            barrier.wait()
        except BaseException:
            barrier.abort()
            for proc in procs:
                proc.terminate()
            raise

    def test_thread(self):
        pid = os.getpid()
        thread = self.prog.thread(pid)
        self.assertEqual(thread.tid, pid)
        self.assertEqual(thread.object, find_task(self.prog, pid))

    def test_main_thread(self):
        self.assertRaisesRegex(
            ValueError,
            "main thread is not defined for the Linux kernel",
            self.prog.main_thread,
        )

    def test_crashed_thread(self):
        self.assertRaisesRegex(
            ValueError,
            "crashed thread is only defined for core dumps",
            self.prog.crashed_thread,
        )

    def test_thread_name(self):
        pid = os.getpid()
        with open(f"/proc/{pid}/comm", "r") as f:
            comm = f.read().strip()
        thread = self.prog.thread(pid)
        self.assertEqual(comm, thread.name)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import contextlib
import os

from drgn import NULL, cast
from drgn.helpers.linux.fs import fget
from drgn.helpers.linux.kernfs import (
    kernfs_children,
    kernfs_name,
    kernfs_parent,
    kernfs_path,
    kernfs_root,
    kernfs_walk,
)
from drgn.helpers.linux.pid import find_task
from tests.linux_kernel import LinuxKernelTestCase


class TestKernfs(LinuxKernelTestCase):
    @classmethod
    def kernfs_node_from_fd(cls, fd):
        file = fget(find_task(cls.prog, os.getpid()), fd)
        return cast("struct kernfs_node *", file.f_inode.i_private)

    def test_kernfs_parent(self):
        with contextlib.ExitStack() as exit_stack:
            fd = os.open("/sys/kernel/vmcoreinfo", os.O_RDONLY)
            exit_stack.callback(os.close, fd)
            dfd = os.open("/sys/kernel", os.O_RDONLY)
            exit_stack.callback(os.close, dfd)
            self.assertEqual(
                kernfs_parent(self.kernfs_node_from_fd(fd)),
                self.kernfs_node_from_fd(dfd),
            )

    def test_kernfs_root(self):
        # Since Linux kernel commit 6138f3451516, "sysfs_root" is ambiguous.
        # Ensure we resolve the symbol from sysfs (mount.c).
        expected_root = self.prog.object("sysfs_root", filename="mount.c")
        for path in ("/sys", "/sys/kernel", "/sys/kernel/vmcoreinfo"):
            with self.subTest(path=path):
                fd = os.open(path, os.O_RDONLY)
                try:
                    self.assertEqual(
                        kernfs_root(self.kernfs_node_from_fd(fd)),
                        expected_root,
                    )
                finally:
                    os.close(fd)

    def test_kernfs_name(self):
        with open("/sys/kernel/vmcoreinfo", "r") as f:
            kn = self.kernfs_node_from_fd(f.fileno())
            self.assertEqual(kernfs_name(kn), b"vmcoreinfo")

    def test_kernfs_path(self):
        with open("/sys/kernel/vmcoreinfo", "r") as f:
            kn = self.kernfs_node_from_fd(f.fileno())
            self.assertEqual(kernfs_path(kn), b"/kernel/vmcoreinfo")

    def test_kernfs_walk(self):
        fds = []
        try:
            for path in ("/sys", "/sys/kernel", "/sys/kernel/vmcoreinfo"):
                fds.append(os.open(path, os.O_RDONLY))
            kns = [self.kernfs_node_from_fd(fd) for fd in fds]
            self.assertEqual(kernfs_walk(kns[0], ""), kns[0])
            self.assertEqual(kernfs_walk(kns[0], "/"), kns[0])
            self.assertEqual(kernfs_walk(kns[0], "kernel"), kns[1])
            self.assertEqual(kernfs_walk(kns[0], "kernel/vmcoreinfo"), kns[2])
            self.assertEqual(
                kernfs_walk(kns[0], "kernel/foobar"),
                NULL(self.prog, "struct kernfs_node *"),
            )
            self.assertEqual(
                kernfs_walk(kns[0], "kernel/foo/bar"),
                NULL(self.prog, "struct kernfs_node *"),
            )
        finally:
            for fd in fds:
                os.close(fd)

    def test_kernfs_walk_follow_symlinks(self):
        fds = []
        path = "/sys/block"
        if not os.path.exists(path):
            self.skipTest(f"{path} does not exist")
        try:
            fd = os.open(path, os.O_RDONLY)
            fds.append(fd)
            kn = self.kernfs_node_from_fd(fd)
            with os.scandir(path) as entries:
                entry = next((e for e in entries if e.is_symlink()), None)
                if not entry:
                    self.skipTest(f"No symlink entries found under {path}")
                sub_path = os.path.join(entry.path, "subsystem")
                fd = os.open(sub_path, os.O_RDONLY)
                fds.append(fd)
                target_kn = self.kernfs_node_from_fd(fd)
                rel_path = os.path.join(entry.name, "subsystem")
                self.assertEqual(
                    kernfs_walk(kn, rel_path),
                    target_kn,
                )
                self.assertNotEqual(
                    kernfs_walk(kn, rel_path, False),
                    target_kn,
                )
        finally:
            for fd in fds:
                os.close(fd)

    def test_kernfs_children(self):
        path = b"/sys/kernel"
        fd = os.open(path, os.O_RDONLY)
        try:
            parent_dev = os.fstat(fd).st_dev
            kn = self.kernfs_node_from_fd(fd)
            self.assertCountEqual(
                [kernfs_name(child) for child in kernfs_children(kn)], os.listdir(path)
            )

            for child in kernfs_children(kn):
                child_fd = os.open(
                    b"/sys/" + kernfs_path(child), os.O_PATH | os.O_NOFOLLOW
                )
                try:
                    if os.fstat(child_fd).st_dev == parent_dev:
                        self.assertEqual(child, self.kernfs_node_from_fd(child_fd))
                finally:
                    os.close(child_fd)
        finally:
            os.close(fd)

        # Check that calling kernfs_children() on a non-directory raises an
        # exception
        path = b"/sys/kernel/vmcoreinfo"
        fd = os.open(path, os.O_RDONLY)
        try:
            kn = self.kernfs_node_from_fd(fd)
            with self.assertRaises(ValueError) as context:
                kernfs_children(kn)
            self.assertEqual(str(context.exception), "not a directory")
        finally:
            os.close(fd)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import functools
import re
import unittest.mock

from drgn import NULL, Object, ObjectNotFoundError
import drgn.helpers.linux.mmzone
from drgn.helpers.linux.mmzone import (
    NODE_DATA,
    _section_flags,
    decode_section_flags,
    for_each_online_pgdat,
    for_each_present_section,
    high_wmark_pages,
    low_wmark_pages,
    min_wmark_pages,
    nr_to_section,
    pfn_to_section,
    pfn_to_section_nr,
    section_decode_mem_map,
    section_mem_map_addr,
    section_nr_to_pfn,
)
from tests.linux_kernel import LinuxKernelTestCase, skip_unless_have_test_kmod


def skip_unless_sparsemem(f):
    @functools.wraps(f)
    def wrapper(self, *args, **kwargs):
        if "mem_section" not in self.prog:
            self.skipTest("kernel does not use SPARSEMEM")

    return wrapper


class TestMmzone(LinuxKernelTestCase):
    @skip_unless_have_test_kmod
    def test_NODE_DATA(self):
        self.assertEqual(
            NODE_DATA(self.prog["drgn_test_nid"]), self.prog["drgn_test_pgdat"]
        )

    @skip_unless_have_test_kmod
    def test_for_each_online_pgdat(self):
        self.assertEqual(
            next(for_each_online_pgdat(self.prog)), self.prog["drgn_test_pgdat"]
        )

    def test_wmark_pages(self):
        pgdat = next(for_each_online_pgdat(self.prog))
        nid = pgdat.node_id.value_()
        zone = pgdat.node_zones + 0
        zone_name = zone.name.string_().decode()
        expected = {}
        with open("/proc/zoneinfo", "r") as f:
            found_zone = False
            for line in f:
                match = re.match(r"Node ([0-9]+), zone\s+(\w+)", line)
                if match:
                    if found_zone:
                        break
                    elif int(match.group(1)) == nid and match.group(2) == zone_name:
                        found_zone = True
                elif found_zone:
                    match = re.match(r"\s*(min|low|high)\s+([0-9]+)", line)
                    if match:
                        expected[match.group(1)] = int(match.group(2))
            else:
                self.fail("zone not found")

        self.assertEqual(min_wmark_pages(zone), expected["min"])
        self.assertEqual(low_wmark_pages(zone), expected["low"])
        self.assertEqual(high_wmark_pages(zone), expected["high"])

    @skip_unless_have_test_kmod
    @skip_unless_sparsemem
    def test_nr_to_section(self):
        self.assertEqual(
            nr_to_section(self.prog["drgn_test_section_nr"]),
            self.prog["drgn_test_mem_section"],
        )

    @skip_unless_sparsemem
    def test_nr_to_section_out_of_bounds(self):
        self.assertEqual(
            nr_to_section(Object(self.prog, "unsigned long", -1)),
            NULL(self.prog, "struct mem_section *"),
        )

    @skip_unless_have_test_kmod
    @skip_unless_sparsemem
    def test_pfn_to_section_nr(self):
        self.assertEqual(
            pfn_to_section_nr(self.prog["drgn_test_pfn"]),
            self.prog["drgn_test_section_nr"],
        )

    @skip_unless_have_test_kmod
    @skip_unless_sparsemem
    def test_section_nr_to_pfn(self):
        self.assertEqual(
            section_nr_to_pfn(self.prog["drgn_test_section_nr"]),
            self.prog["drgn_test_section_pfn"],
        )

    @skip_unless_have_test_kmod
    @skip_unless_sparsemem
    def test_pfn_to_section(self):
        self.assertEqual(
            pfn_to_section(self.prog["drgn_test_pfn"]),
            self.prog["drgn_test_mem_section"],
        )

    @skip_unless_have_test_kmod
    @skip_unless_sparsemem
    def test_section_mem_map_addr(self):
        self.assertEqual(
            section_mem_map_addr(self.prog["drgn_test_mem_section"]),
            self.prog["drgn_test_section_mem_map"],
        )

    @skip_unless_have_test_kmod
    @skip_unless_sparsemem
    def test_section_decode_mem_map(self):
        self.assertEqual(
            section_decode_mem_map(self.prog["drgn_test_mem_section"]),
            self.prog["drgn_test_section_decoded_mem_map"],
        )

    @skip_unless_have_test_kmod
    @skip_unless_sparsemem
    def test_section_flags(self):
        flags = _section_flags(self.prog)
        for name in (
            "SECTION_MARKED_PRESENT",
            "SECTION_HAS_MEM_MAP",
            "SECTION_IS_ONLINE",
            "SECTION_IS_EARLY",
            "SECTION_TAINT_ZONE_DEVICE",
        ):
            with self.subTest(flag=name):
                try:
                    expected = self.prog["drgn_test_" + name].value_()
                except ObjectNotFoundError:
                    self.skipTest(f"{name} is not defined")
                self.assertEqual(flags[name], expected)

    @skip_unless_have_test_kmod
    @skip_unless_sparsemem
    def test_section_flag_getters(self):
        # These helpers are tricky to test because a typical VM setup will only
        # have sections with all of these flags set. So, we create fake
        # mem_section structures in the test kmod and check those. But that
        # means we have to monkey patch nr_to_section() to test the
        # foo_section_nr() helpers.
        prev_section = self.prog["drgn_test_valid_section"]
        for name in (
            "present_section",
            "valid_section",
            "online_section",
            "early_section",
        ):
            with self.subTest(helper=name):
                try:
                    section = self.prog["drgn_test_" + name]
                except ObjectNotFoundError:
                    self.skipTest(f"{name} is not defined")

                try:
                    helper = getattr(drgn.helpers.linux.mmzone, name)
                    self.assertTrue(helper(section))
                    self.assertFalse(helper(prev_section))
                    self.assertFalse(helper(NULL(self.prog, "struct mem_section *")))

                    nr_helper = getattr(drgn.helpers.linux.mmzone, name + "_nr")
                    with unittest.mock.patch(
                        "drgn.helpers.linux.mmzone.nr_to_section"
                    ) as mock_nr_to_section:
                        mock_nr_to_section.return_value = section
                        self.assertTrue(nr_helper(self.prog, 0))
                        mock_nr_to_section.return_value = prev_section
                        self.assertFalse(nr_helper(self.prog, 0))
                finally:
                    prev_section = section

    @skip_unless_have_test_kmod
    @skip_unless_sparsemem
    def test_decode_section_flags(self):
        self.assertIn(
            "SECTION_MARKED_PRESENT",
            decode_section_flags(self.prog["drgn_test_mem_section"]).split("|"),
        )

    @skip_unless_have_test_kmod
    @skip_unless_sparsemem
    def test_for_each_present_section(self):
        present_sections = self.prog["drgn_test_present_sections"]
        self.assertEqual(
            list(for_each_present_section(self.prog)),
            [
                (present_sections[i].nr.value_(), present_sections[i].section.read_())
                for i in range(self.prog["drgn_test_num_present_sections"])
            ],
        )



#!/usr/bin/env python3
import contextlib
import os
from pathlib import Path
import tempfile

import drgn.commands._builtin  # noqa: F401
from tests.commands import CommandTestCase


@contextlib.contextmanager
def temporary_working_directory():
    old_working_directory = os.getcwd()
    with tempfile.TemporaryDirectory() as f:
        try:
            os.chdir(f)
            yield f
        finally:
            os.chdir(old_working_directory)


class RedirectedFile:
    def __init__(self, f):
        self.tempfile = f
        self.value = None


@contextlib.contextmanager
def redirect(stdout=False, stderr=False):
    # To redirect stdout for commands, we need a real file descriptor, not just
    # a StringIO
    with contextlib.ExitStack() as stack:
        f = stack.enter_context(tempfile.TemporaryFile("w+t"))
        if stdout:
            stack.enter_context(contextlib.redirect_stdout(f))
        if stderr:
            stack.enter_context(contextlib.redirect_stderr(f))
        redir = RedirectedFile(f)
        try:
            yield redir
        finally:
            f.seek(0)
            redir.value = f.read()


class TestPyCommand(CommandTestCase):

    def test_py_redirect(self):
        with temporary_working_directory() as temp_dir:
            path = Path(temp_dir) / "6"
            self.run_command("py var = 5; var > 6")
            self.assertEqual(path.read_text(), "5\n")

    def test_py_paren_avoid_redirect(self):
        self.run_command("py var = 5; (var > 6)")
        with redirect(stdout=True) as f:
            self.run_command("py var = 5; (var > 6)")
        self.assertEqual(f.value, "False\n")

    def test_py_pipe(self):
        with redirect(stdout=True) as f:
            self.run_command("py echo = 5; 2 | echo + 5")
        self.assertEqual(f.value, "+ 5\n")

    def test_py_avoid_pipe(self):
        with redirect(stdout=True) as f:
            self.run_command("py echo = 5; (2 | (echo + 5))")
        self.assertEqual(f.value, "10\n")

    def test_py_chooses_first_pipe(self):
        with redirect(stdout=True) as f:
            # If the first | is used to separate the Python from the pipeline
            # (the expected behavior), then we'll get the value 5 written into
            # the "echo" command, which will ignore that and write "+ 6" through
            # the cat process to stdout. If the second | is used to separate the
            # Python from the pipeline, then we'll get "15" written into the cat
            # process. If none of the | were interpreted as a pipeline operator,
            # then the statement would output 31.
            self.run_command("py echo = 5; cat = 16; 5 | echo + 6 | cat")
        self.assertEqual("+ 6\n", f.value)

    def test_py_traceback_on_syntax_error(self):
        with redirect(stderr=True) as f:
            self.run_command("py a +")
        # SyntaxError does not print the "Traceback" header. Rather than trying
        # to assert too much about the format of the traceback, just assert that
        # the incorrect code is shown, as it would be for a traceback.
        self.assertTrue("a +" in f.value)
        self.assertTrue("SyntaxError" in f.value)

    def test_py_traceback_on_exception(self):
        with redirect(stderr=True) as f:
            self.run_command("py raise Exception('text')")
        self.assertTrue(f.value.startswith("Traceback"))
        self.assertTrue("Exception" in f.value)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

from typing import Optional


def dot_join(*args: Optional[str]) -> str:
    return ".".join([s for s in args if s])



# Copyright (c) 2025 Oracle and/or its affiliates
# SPDX-License-Identifier: LGPL-2.1-or-later
import argparse
import os
from pathlib import Path
import shlex
import subprocess
import sys
from typing import List

from vmtest.config import ARCHITECTURES, HOST_ARCHITECTURE

_PASSTHROUGH_ENV_VARS = (
    "TERM",
    "COLORTERM",
    "http_proxy",
    "https_proxy",
    "ftp_proxy",
    "no_proxy",
)


# When entering a chroot, we want a clean environment. These functions preserve
# only a few specific environment variables and execute a login shell, which
# will define anything else important.
def chroot_sh_args(new_root: str) -> List[str]:
    return [
        "chroot",
        new_root,
        "env",
        "-i",
        *(
            f"{var}={value}"
            for var in _PASSTHROUGH_ENV_VARS
            if (value := os.getenv(var)) is not None
        ),
        "/bin/sh",
        "-l",
        "-c",
    ]


def chroot_sh_cmd(new_root: str) -> str:
    return " ".join(
        [
            "chroot",
            new_root,
            "env",
            "-i",
            *(f'${{{var}+{var}="${var}"}}' for var in _PASSTHROUGH_ENV_VARS),
            "/bin/sh",
            "-l",
            "-c",
        ]
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="run commands in the root filesystems for vmtest",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "-d",
        "--directory",
        metavar="DIR",
        type=Path,
        default="build/vmtest",
        help="directory for vmtest artifacts",
    )
    parser.add_argument(
        "-a",
        "--architecture",
        type=str,
        choices=sorted(ARCHITECTURES),
        default=None if HOST_ARCHITECTURE is None else HOST_ARCHITECTURE.name,
        required=HOST_ARCHITECTURE is None,
        help="architecture to run in",
    )
    parser.add_argument(
        "command",
        type=str,
        nargs=argparse.REMAINDER,
        help="command to run in rootfs (default: bash -i)",
    )
    args = parser.parse_args()
    arch = ARCHITECTURES[args.architecture]
    dir = args.directory / arch.name / "rootfs"
    command = (
        " ".join([shlex.quote(arg) for arg in args.command])
        if args.command
        else "bash -i"
    )
    sys.exit(
        subprocess.run(
            [
                "unshare",
                "--map-root-user",
                "--map-auto",
                "--fork",
                "--pid",
                f"--mount-proc={dir / 'proc'}",
                *chroot_sh_args(dir),
                command,
            ],
        ).returncode
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: GPL-3.0-or-later

import argparse
import asyncio
from collections import defaultdict
import logging
import operator
import os
from pathlib import Path
import re
import shutil
import sys
from typing import Any, Awaitable, Callable, Dict, Iterable, List, Tuple, TypeVar

import aiohttp

from util import KernelVersion
from vmtest.download import VMTEST_GITHUB_RELEASE
from vmtest.githubapi import AioGitHubApi

logger = logging.getLogger(__name__)


def _prompt_yes_no(prompt: str, default: bool = True) -> bool:
    prompt += " [Y/n] " if default else " [y/N] "
    sys.stdout.flush()
    sys.stderr.write(prompt)
    sys.stderr.flush()
    answer = input().strip().lower()
    if answer.startswith("y"):
        return True
    elif answer.startswith("n"):
        return False
    else:
        return default


def _kernel_asset_sort_key(x: Tuple[str, str, Any]) -> Tuple[KernelVersion, str]:
    release, arch, _ = x
    return KernelVersion(release), arch


_AssetT = TypeVar("_AssetT")


async def _prune(
    *,
    kernels: Iterable[Tuple[str, str, _AssetT]],
    kmods: Iterable[Tuple[str, str, _AssetT]] = (),
    keep: int,
    asset_name: Callable[[_AssetT], str],
    delete_asset: Callable[[_AssetT], Awaitable[None]],
) -> int:
    def print_asset_list(assets: List[Tuple[str, str, _AssetT]]) -> None:
        for _, _, asset in assets:
            print(f"    {asset_name(asset)}")

    grouped_kernels: Dict[Tuple[str, str, str], List[Tuple[str, _AssetT]]] = (
        defaultdict(list)
    )
    unrecognized_kernels: List[Tuple[str, str, _AssetT]] = []

    for release, arch, asset in kernels:
        match = re.fullmatch(
            r"(?P<major_minor>[0-9]+\.[0-9]+)\.(?P<patch_rc>[0-9]+(?:-rc[0-9]+)?)-vmtest(?P<vmtest>[0-9]+(?:\.[0-9]+)*)(?P<flavor>[^.]+)",
            release,
        )
        if match:
            grouped_kernels[
                (match.group("major_minor"), match.group("flavor"), arch)
            ].append((release, asset))
        else:
            unrecognized_kernels.append((release, arch, asset))

    kernels_to_keep = {(release, arch) for release, arch, _ in unrecognized_kernels}
    kernels_to_delete: List[Tuple[str, str, _AssetT]] = []
    for (_, _, arch), kernel_group in grouped_kernels.items():
        kernel_group.sort(key=lambda x: KernelVersion(x[0]), reverse=True)
        for release, _ in kernel_group[:keep]:
            kernels_to_keep.add((release, arch))
        kernels_to_delete.extend(
            (release, arch, asset) for release, asset in kernel_group[keep:]
        )
    kernels_to_delete.sort(key=_kernel_asset_sort_key)

    kmods_to_delete = [
        (release, arch, asset)
        for release, arch, asset in kmods
        if (release, arch) not in kernels_to_keep
    ]
    kmods_to_delete.sort(key=_kernel_asset_sort_key)

    if kernels_to_delete:
        print("Deleting kernels:")
        print_asset_list(kernels_to_delete)

    if kmods_to_delete:
        print("Deleting kernel modules:")
        print_asset_list(kmods_to_delete)

    if unrecognized_kernels:
        unrecognized_kernels.sort(key=_kernel_asset_sort_key)
        print("Ignoring unrecognized kernels:")
        print_asset_list(unrecognized_kernels)

    if not kernels_to_delete and not kmods_to_delete:
        return 0

    if not _prompt_yes_no("Continue?", default=False):
        return 1

    to_delete = kernels_to_delete + kmods_to_delete
    for _, _, asset in to_delete:
        await delete_asset(asset)
    return 0


async def _delete_local_asset(asset: Path) -> None:
    try:
        asset.unlink()
    except IsADirectoryError:
        shutil.rmtree(asset)


async def main() -> None:
    parser = argparse.ArgumentParser(
        description="delete old vmtest kernels",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--directory",
        metavar="DIR",
        type=Path,
        default="build/vmtest",
        help="directory for vmtest artifacts",
    )
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "-l",
        "--local",
        action="store_true",
        default=argparse.SUPPRESS,
        help="delete locally downloaded vmtest kernels and their kernel modules",
    )
    group.add_argument(
        "-r",
        "--remote",
        dest="local",
        action="store_false",
        default=argparse.SUPPRESS,
        help="delete uploaded vmtest kernel assets",
    )
    parser.add_argument(
        "-k", "--keep", type=int, default=1, help="number of latest kernels to keep"
    )
    args = parser.parse_args()

    if args.keep <= 0:
        sys.exit("keep must be >= 1")

    if args.local:
        status = await _prune(
            kernels=(
                (kernel.name[len("kernel-") :], kernel.parent.name, kernel)
                for kernel in args.directory.glob("*/kernel-*")
            ),
            kmods=(
                (kmod.name[len("drgn_test-") : -len(".ko")], kmod.parent.name, kmod)
                for kmod in args.directory.glob("*/drgn_test-*.ko")
            ),
            keep=args.keep,
            asset_name=str,
            delete_asset=_delete_local_asset,
        )
    else:
        GITHUB_TOKEN = os.getenv("GITHUB_TOKEN")
        if GITHUB_TOKEN is None:
            sys.exit("GITHUB_TOKEN environment variable is not set")
        async with aiohttp.ClientSession(trust_env=True) as session:
            gh = AioGitHubApi(session, GITHUB_TOKEN)

            github_release = await gh.get_release_by_tag(
                *VMTEST_GITHUB_RELEASE, cache=args.directory / "github_release.json"
            )

            kernels = []
            for asset in github_release["assets"]:
                match = re.fullmatch(
                    r"kernel-(?P<release>.*)\.(?P<arch>[^.]+)\.tar\.zst", asset["name"]
                )
                if match:
                    kernels.append((match.group("release"), match.group("arch"), asset))
                elif asset["name"].startswith("kernel-"):
                    logger.warning("ignoring unrecognized kernel: %s", asset["name"])

            async def _delete_remote_asset(asset: Dict[str, Any]) -> None:
                await gh.delete(asset["url"])

            status = await _prune(
                kernels=kernels,
                keep=args.keep,
                asset_name=operator.itemgetter("name"),
                delete_asset=_delete_remote_asset,
            )

    sys.exit(status)


if __name__ == "__main__":
    asyncio.run(main())



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import argparse
import logging
import os
from pathlib import Path
import shutil
import sys
import tempfile
import urllib.request

import uritemplate

from vmtest.config import ARCHITECTURES, KERNEL_ORG_COMPILER_VERSION, compiler_name
from vmtest.githubapi import GitHubApi

logger = logging.getLogger(__name__)


_KERNEL_ORG_COMPILER_URL = "https://mirrors.kernel.org/pub/tools/crosstool/"


def _kernel_org_compiler_url(host_name: str, target_name: str) -> str:
    return f"{_KERNEL_ORG_COMPILER_URL}files/bin/{host_name}/{KERNEL_ORG_COMPILER_VERSION}/{compiler_name(host_name, target_name)}.tar.xz"


def main() -> None:
    logging.basicConfig(
        format="%(asctime)s:%(levelname)s:%(name)s:%(message)s", level=logging.INFO
    )

    parser = argparse.ArgumentParser(
        description="mirror compilers from kernel.org to GitHub"
    )
    parser.parse_args()

    GITHUB_TOKEN = os.getenv("GITHUB_TOKEN")
    if GITHUB_TOKEN is None:
        sys.exit("GITHUB_TOKEN environment variable is not set")
    gh = GitHubApi(GITHUB_TOKEN)

    host_names = []
    target_names = []
    for arch in ARCHITECTURES.values():
        if arch.kernel_org_compiler_host_name is not None:
            host_names.append(arch.kernel_org_compiler_host_name)
        target_names.append(arch.kernel_org_compiler_target_name)

    release = gh.get_release_by_tag("osandov", "drgn", "vmtest-compilers")
    upload_url = release["upload_url"]
    available_compilers = {asset["name"] for asset in release["assets"]}

    with tempfile.TemporaryDirectory() as tmp:
        tmp_dir = Path(tmp)

        for host_name in host_names:
            for target_name in target_names:
                url = _kernel_org_compiler_url(host_name, target_name)
                file_name = url.rpartition("/")[2]

                if file_name in available_compilers:
                    logger.info("%s is already available", file_name)
                    continue

                file_path = tmp_dir / file_name

                with file_path.open("w+b") as f:
                    logger.info("downloading %s", url)
                    with urllib.request.urlopen(url) as resp:
                        shutil.copyfileobj(resp, f)
                    logger.info("downloaded %s", url)

                    content_length = f.tell()
                    f.flush()
                    f.seek(0)

                    logger.info("uploading %s", file_name)
                    print(uritemplate.expand(upload_url, name=file_name))
                    gh.upload(
                        uritemplate.expand(upload_url, name=file_name),
                        f,
                        "application/x-xz",
                        content_length,
                    )
                    logger.info("uploaded %s", file_name)

                file_path.unlink()


if __name__ == "__main__":
    main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

from drgn import Language, Qualifiers, TypeEnumerator, TypeMember
from tests import MockProgramTestCase


class TestFormatTypeName(MockProgramTestCase):
    def setUp(self):
        super().setUp()
        self.prog.language = Language.CPP

    def test_tagged(self):
        for keyword in ("struct", "union", "class"):
            with self.subTest(keyword=keyword):
                constructor = getattr(self.prog, keyword + "_type")
                self.assertEqual(
                    constructor(
                        "foo",
                        4,
                        (TypeMember(self.prog.int_type("int", 4, True), "foo", 0),),
                    ).type_name(),
                    "foo",
                )
        with self.subTest(keyword="enum"):
            self.assertEqual(
                self.prog.enum_type(
                    "foo",
                    self.prog.int_type("int", 4, True),
                    (TypeEnumerator("FOO", 1),),
                ).type_name(),
                "foo",
            )

    def test_qualified_tagged(self):
        for keyword in ("struct", "union", "class"):
            with self.subTest(keyword=keyword):
                constructor = getattr(self.prog, keyword + "_type")
                self.assertEqual(
                    constructor(
                        "foo",
                        4,
                        (TypeMember(self.prog.int_type("int", 4, True), "foo", 0),),
                        qualifiers=Qualifiers.CONST,
                    ).type_name(),
                    "const foo",
                )
        with self.subTest(keyword="enum"):
            self.assertEqual(
                self.prog.enum_type(
                    "foo",
                    self.prog.int_type("int", 4, True),
                    (TypeEnumerator("FOO", 1),),
                    qualifiers=Qualifiers.CONST,
                ).type_name(),
                "const foo",
            )

    def test_anonymous_tagged(self):
        for keyword in ("struct", "union", "class"):
            with self.subTest(keyword=keyword):
                constructor = getattr(self.prog, keyword + "_type")
                self.assertEqual(
                    constructor(
                        None,
                        4,
                        (TypeMember(self.prog.int_type("int", 4, True), "foo", 0),),
                    ).type_name(),
                    keyword + " <anonymous>",
                )
        with self.subTest(keyword="enum"):
            self.assertEqual(
                self.prog.enum_type(
                    None,
                    self.prog.int_type("int", 4, True),
                    (TypeEnumerator("FOO", 1),),
                ).type_name(),
                "enum <anonymous>",
            )

    def test_qualified_anonymous_tagged(self):
        for keyword in ("struct", "union", "class"):
            with self.subTest(keyword=keyword):
                constructor = getattr(self.prog, keyword + "_type")
                self.assertEqual(
                    constructor(
                        None,
                        4,
                        (TypeMember(self.prog.int_type("int", 4, True), "foo", 0),),
                        qualifiers=Qualifiers.CONST,
                    ).type_name(),
                    "const " + keyword + " <anonymous>",
                )
        with self.subTest(keyword="enum"):
            self.assertEqual(
                self.prog.enum_type(
                    None,
                    self.prog.int_type("int", 4, True),
                    (TypeEnumerator("FOO", 1),),
                    qualifiers=Qualifiers.CONST,
                ).type_name(),
                "const enum <anonymous>",
            )


class TestFormatType(MockProgramTestCase):
    def setUp(self):
        super().setUp()
        self.prog.language = Language.CPP

    def test_tagged(self):
        for keyword in ("struct", "union", "class"):
            with self.subTest(keyword=keyword):
                constructor = getattr(self.prog, keyword + "_type")
                self.assertEqual(
                    str(
                        constructor(
                            "foo",
                            4,
                            (TypeMember(self.prog.int_type("int", 4, True), "foo", 0),),
                        )
                    ),
                    f"""\
{keyword} foo {{
	int foo;
}}""",
                )
        with self.subTest(keyword="enum"):
            self.assertEqual(
                str(
                    self.prog.enum_type(
                        "foo",
                        self.prog.int_type("int", 4, True),
                        (TypeEnumerator("FOO", 1),),
                    )
                ),
                """\
enum foo {
	FOO = 1,
}""",
            )

    def test_anonymous_tagged(self):
        for keyword in ("struct", "union", "class"):
            with self.subTest(keyword=keyword):
                constructor = getattr(self.prog, keyword + "_type")
                self.assertEqual(
                    str(
                        constructor(
                            None,
                            4,
                            (TypeMember(self.prog.int_type("int", 4, True), "foo", 0),),
                        )
                    ),
                    f"""\
{keyword} {{
	int foo;
}}""",
                )
        with self.subTest(keyword="enum"):
            self.assertEqual(
                str(
                    self.prog.enum_type(
                        None,
                        self.prog.int_type("int", 4, True),
                        (TypeEnumerator("FOO", 1),),
                    )
                ),
                """\
enum {
	FOO = 1,
}""",
            )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import errno
import os
import re

from drgn.helpers.linux.printk import PrintkRecord, get_printk_records
from tests.linux_kernel import LinuxKernelTestCase


def unescape_text(s):
    return re.sub(
        rb"\\x([0-9A-Fa-f]{2})", lambda match: bytes([int(match.group(1), 16)]), s
    )


def get_kmsg_records():
    result = []
    with open("/dev/kmsg", "rb") as f:
        fd = f.fileno()
        os.set_blocking(fd, False)
        while True:
            try:
                record = os.read(fd, 4096)
            except OSError as e:
                if e.errno == errno.EAGAIN:
                    break
                else:
                    raise
            prefix, _, escaped = record.partition(b";")
            fields = prefix.split(b",")
            syslog = int(fields[0])
            escaped_lines = escaped.splitlines()
            context = {}
            for escaped_line in escaped_lines[1:]:
                assert escaped_line.startswith(b" ")
                key, value = escaped_line[1:].split(b"=", 1)
                context[unescape_text(key)] = unescape_text(value)
            result.append(
                PrintkRecord(
                    text=unescape_text(escaped_lines[0]),
                    facility=syslog >> 3,
                    level=syslog & 7,
                    seq=int(fields[1]),
                    timestamp=int(fields[2]) * 1000,
                    caller_tid=None,
                    caller_cpu=None,
                    continuation=b"c" in fields[3],
                    context=context,
                )
            )
    return result


class TestPrintk(LinuxKernelTestCase):
    def test_get_printk_records(self):
        self.assertEqual(
            get_kmsg_records(),
            [
                record._replace(
                    # Round timestamp down since /dev/kmsg only has microsecond
                    # granularity.
                    timestamp=record.timestamp // 1000 * 1000,
                    # Remove caller ID since it's only available from /dev/kmsg
                    # when the kernel is compiled with CONFIG_PRINTK_CALLER.
                    caller_tid=None,
                    caller_cpu=None,
                )
                for record in get_printk_records(self.prog)
            ],
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later
import ctypes
import itertools
import os
import sys
import tempfile
import unittest.mock

from _drgn_util.elf import ET, PT
from drgn import (
    Architecture,
    FaultError,
    FindObjectFlags,
    Language,
    NoDefaultProgramError,
    Object,
    ObjectNotFoundError,
    Platform,
    PlatformFlags,
    Program,
    ProgramFlags,
    Qualifiers,
    TypeKind,
    TypeMember,
    get_default_prog,
    host_platform,
    set_default_prog,
)
from tests import (
    DEFAULT_LANGUAGE,
    MOCK_32BIT_PLATFORM,
    MOCK_PLATFORM,
    MockMemorySegment,
    MockObject,
    MockProgramTestCase,
    TestCase,
    mock_program,
)
from tests.elfwriter import ElfSection, create_elf_file


def zero_memory_read(address, count, offset, physical):
    return bytes(count)


class TestProgram(TestCase):
    def test_default_program(self):
        self.assertRaises(NoDefaultProgramError, get_default_prog)
        prog = Program()
        prog2 = Program()
        try:
            set_default_prog(prog)
            self.assertIs(get_default_prog(), prog)
            set_default_prog(prog2)
            self.assertIs(get_default_prog(), prog2)
        finally:
            set_default_prog(None)
        self.assertRaises(NoDefaultProgramError, get_default_prog)

    def test_default_program_reference_counting(self):
        try:
            set_default_prog(Program())
            self.assertGreater(sys.getrefcount(get_default_prog()), 1)
        finally:
            set_default_prog(None)

    def test_set_pid(self):
        # Debug the running Python interpreter itself.
        prog = Program()
        self.assertIsNone(prog.platform)
        self.assertFalse(prog.flags & ProgramFlags.IS_LIVE)
        prog.set_pid(os.getpid())
        self.assertIsNone(prog.core_dump_path)
        self.assertEqual(prog.platform, host_platform)
        self.assertTrue(prog.flags & ProgramFlags.IS_LIVE)
        self.assertRaisesRegex(
            ValueError,
            "program memory was already initialized",
            prog.set_pid,
            os.getpid(),
        )

    def test_pid_memory(self):
        data = b"hello, world!"
        buf = ctypes.create_string_buffer(data)
        address = ctypes.addressof(buf)

        # QEMU user-mode emulation doesn't seem to emulate /proc/$pid/mem
        # correctly on a 64-bit host with a 32-bit guest; see
        # https://gitlab.com/qemu-project/qemu/-/issues/698. Packit uses mock
        # to cross-compile and test packages, which in turn uses QEMU user-mode
        # emulation. Skip this test if /proc/$pid/mem doesn't work so that
        # those builds succeed.
        try:
            with open("/proc/self/mem", "rb") as f:
                f.seek(address)
                functional_proc_pid_mem = f.read(len(data)) == data
        except OSError:
            functional_proc_pid_mem = False
        if not functional_proc_pid_mem:
            self.skipTest("/proc/$pid/mem is not functional")

        prog = Program()
        prog.set_pid(os.getpid())

        self.assertEqual(prog.read(ctypes.addressof(buf), len(data)), data)

    def test_object_not_found_error(self):
        prog = mock_program()

        with self.assertRaisesRegex(
            ObjectNotFoundError, "^could not find constant 'foo'$"
        ) as cm:
            prog.constant("foo")
        self.assertEqual(cm.exception.name, "foo")

        with self.assertRaisesRegex(
            ObjectNotFoundError, "^could not find constant 'foo' in 'foo.c'$"
        ) as cm:
            prog.constant("foo", "foo.c")
        self.assertEqual(cm.exception.name, "foo")

        with self.assertRaisesRegex(
            ObjectNotFoundError, "^could not find function 'foo'$"
        ) as cm:
            prog.function("foo")
        self.assertEqual(cm.exception.name, "foo")

        with self.assertRaisesRegex(
            ObjectNotFoundError, "^could not find function 'foo' in 'foo.c'$"
        ) as cm:
            prog.function("foo", "foo.c")
        self.assertEqual(cm.exception.name, "foo")

        with self.assertRaisesRegex(
            ObjectNotFoundError, "^could not find variable 'foo'$"
        ) as cm:
            prog.variable("foo")
        self.assertEqual(cm.exception.name, "foo")

        with self.assertRaisesRegex(
            ObjectNotFoundError, "^could not find variable 'foo' in 'foo.c'$"
        ) as cm:
            prog.variable("foo", "foo.c")
        self.assertEqual(cm.exception.name, "foo")

        with self.assertRaisesRegex(
            ObjectNotFoundError, "^could not find 'foo'$"
        ) as cm:
            prog["foo"]
        self.assertEqual(cm.exception.name, "foo")

        # If name isn't a string, prog.object(name) should raise TypeError, and
        # prog[name] should raise KeyError (not ObjectNotFoundError).
        self.assertRaises(TypeError, prog.object, 9)
        with self.assertRaises(KeyError) as cm:
            prog[9]
        self.assertIs(type(cm.exception), KeyError)

    def test_type_lookup_error(self):
        prog = mock_program()

        self.assertRaisesRegex(LookupError, "^could not find 'foo'$", prog.type, "foo")
        self.assertRaisesRegex(
            LookupError, "^could not find 'foo' in 'foo.c'$", prog.type, "foo", "foo.c"
        )

    def test_flags(self):
        self.assertIsInstance(mock_program().flags, ProgramFlags)

    def test_debug_info(self):
        Program().load_debug_info([])

    def test_language(self):
        prog = Program()
        self.assertEqual(prog.language, DEFAULT_LANGUAGE)
        prog.language = Language.CPP
        self.assertEqual(prog.language, Language.CPP)
        prog.language = Language.C
        self.assertEqual(prog.language, Language.C)
        self.assertRaisesRegex(
            TypeError, "language must be Language", setattr, prog, "language", "CPP"
        )

    def test_language_del(self):
        with self.assertRaises(AttributeError):
            del Program().language

    def test_address_size(self):
        self.assertEqual(mock_program().address_size(), 8)
        self.assertEqual(mock_program(MOCK_32BIT_PLATFORM).address_size(), 4)
        self.assertRaisesRegex(
            ValueError, "address size is not known", Program().address_size
        )


class TestMemory(TestCase):
    def test_simple_read(self):
        data = b"hello, world"
        prog = mock_program(segments=[MockMemorySegment(data, 0xFFFF0000, 0xA0)])
        self.assertEqual(prog.read(0xFFFF0000, len(data)), data)
        self.assertEqual(prog.read(0xA0, len(data), True), data)

    def test_read_c_string(self):
        data = b"hello, world"
        prog = mock_program(
            segments=[MockMemorySegment(data + b"\0", 0xFFFF0000, 0xA0)]
        )
        self.assertEqual(prog.read_c_string(0xFFFF0000), data)
        self.assertEqual(prog.read_c_string(0xA0, True), data)

    def test_read_c_string_max_size(self):
        for data in (b"hello, world", b"hell", b"he"):
            with self.subTest(data=data):
                prog = mock_program(
                    segments=[
                        MockMemorySegment(
                            data + (b"\0" if len(data) < 4 else b""), 0xFFFF0000, 0xA0
                        )
                    ]
                )
                self.assertEqual(prog.read_c_string(0xFFFF0000, max_size=4), data[:4])
                self.assertEqual(prog.read_c_string(0xA0, True, max_size=4), data[:4])

    def test_read_unsigned(self):
        data = b"\x01\x02\x03\x04\x05\x06\x07\x08"
        for word_size in [8, 4]:
            for byteorder in ["little", "big"]:
                flags = PlatformFlags(0)
                if word_size == 8:
                    flags |= PlatformFlags.IS_64_BIT
                if byteorder == "little":
                    flags |= PlatformFlags.IS_LITTLE_ENDIAN
                prog = mock_program(
                    Platform(Architecture.UNKNOWN, flags),
                    segments=[MockMemorySegment(data, 0xFFFF0000, 0xA0)],
                )
                for size in [1, 2, 4, 8]:
                    read_fn = getattr(prog, f"read_u{8 * size}")
                    value = int.from_bytes(data[:size], byteorder)
                    self.assertEqual(read_fn(0xFFFF0000), value)
                    self.assertEqual(read_fn(0xA0, True), value)
                    if size == word_size:
                        self.assertEqual(prog.read_word(0xFFFF0000), value)
                        self.assertEqual(prog.read_word(0xA0, True), value)

        prog = mock_program(
            MOCK_32BIT_PLATFORM, segments=[MockMemorySegment(data, 0xFFFF0000, 0xA0)]
        )

    def test_bad_address(self):
        data = b"hello, world!"
        prog = mock_program(segments=[MockMemorySegment(data, 0xFFFF0000)])
        self.assertRaisesRegex(
            FaultError, "could not find memory segment", prog.read, 0xDEADBEEF, 4
        )
        self.assertRaisesRegex(
            FaultError,
            "could not find physical memory segment",
            prog.read,
            0xFFFF0000,
            4,
            True,
        )

    def test_segment_overflow(self):
        data = b"hello, world!"
        prog = mock_program(segments=[MockMemorySegment(data, 0xFFFF0000)])
        self.assertRaisesRegex(
            FaultError,
            "could not find memory segment",
            prog.read,
            0xFFFF0000,
            len(data) + 1,
        )

    def test_adjacent_segments(self):
        data = b"hello, world!\0foobar"
        prog = mock_program(
            segments=[
                MockMemorySegment(data[:4], 0xFFFF0000),
                MockMemorySegment(data[4:14], 0xFFFF0004),
                MockMemorySegment(data[14:], 0xFFFFF000),
            ]
        )
        self.assertEqual(prog.read(0xFFFF0000, 14), data[:14])

    def test_address_overflow(self):
        for bits in (64, 32):
            with self.subTest(bits=bits):
                prog = mock_program(
                    segments=[
                        MockMemorySegment(b"cd", 0x0),
                        MockMemorySegment(b"abyz", 2**bits - 2),
                    ],
                    platform=MOCK_PLATFORM if bits == 64 else MOCK_32BIT_PLATFORM,
                )
                for start in range(3):
                    for size in range(4 - start):
                        self.assertEqual(
                            prog.read((2**bits - 2 + start) % 2**64, size),
                            b"abcd"[start : start + size],
                        )

    def test_overlap_same_address_smaller_size(self):
        # Existing segment: |_______|
        # New segment:      |___|
        prog = Program(MOCK_PLATFORM)
        segment1 = unittest.mock.Mock(side_effect=zero_memory_read)
        segment2 = unittest.mock.Mock(side_effect=zero_memory_read)
        prog.add_memory_segment(0xFFFF0000, 128, segment1)
        prog.add_memory_segment(0xFFFF0000, 64, segment2)
        prog.read(0xFFFF0000, 128)
        segment1.assert_called_once_with(0xFFFF0040, 64, 64, False)
        segment2.assert_called_once_with(0xFFFF0000, 64, 0, False)

    def test_overlap_within_segment(self):
        # Existing segment: |_______|
        # New segment:        |___|
        prog = Program(MOCK_PLATFORM)
        segment1 = unittest.mock.Mock(side_effect=zero_memory_read)
        segment2 = unittest.mock.Mock(side_effect=zero_memory_read)
        prog.add_memory_segment(0xFFFF0000, 128, segment1)
        prog.add_memory_segment(0xFFFF0020, 64, segment2)
        prog.read(0xFFFF0000, 128)
        segment1.assert_has_calls(
            [
                unittest.mock.call(0xFFFF0000, 32, 00, False),
                unittest.mock.call(0xFFFF0060, 32, 96, False),
            ]
        )
        segment2.assert_called_once_with(0xFFFF0020, 64, 0, False)

    def test_overlap_same_segment(self):
        # Existing segment: |_______|
        # New segment:      |_______|
        prog = Program(MOCK_PLATFORM)
        segment1 = unittest.mock.Mock(side_effect=zero_memory_read)
        segment2 = unittest.mock.Mock(side_effect=zero_memory_read)
        prog.add_memory_segment(0xFFFF0000, 128, segment1)
        prog.add_memory_segment(0xFFFF0000, 128, segment2)
        prog.read(0xFFFF0000, 128)
        segment1.assert_not_called()
        segment2.assert_called_once_with(0xFFFF0000, 128, 0, False)

    def test_overlap_same_address_larger_size(self):
        # Existing segment: |___|
        # New segment:      |_______|
        prog = Program(MOCK_PLATFORM)
        segment1 = unittest.mock.Mock(side_effect=zero_memory_read)
        segment2 = unittest.mock.Mock(side_effect=zero_memory_read)
        prog.add_memory_segment(0xFFFF0000, 64, segment1)
        prog.add_memory_segment(0xFFFF0000, 128, segment2)
        prog.read(0xFFFF0000, 128)
        segment1.assert_not_called()
        segment2.assert_called_once_with(0xFFFF0000, 128, 0, False)

    def test_overlap_segment_tail(self):
        # Existing segment: |_______|
        # New segment:          |_______|
        prog = Program(MOCK_PLATFORM)
        segment1 = unittest.mock.Mock(side_effect=zero_memory_read)
        segment2 = unittest.mock.Mock(side_effect=zero_memory_read)
        prog.add_memory_segment(0xFFFF0000, 128, segment1)
        prog.add_memory_segment(0xFFFF0040, 128, segment2)
        prog.read(0xFFFF0000, 192)
        segment1.assert_called_once_with(0xFFFF0000, 64, 0, False)
        segment2.assert_called_once_with(0xFFFF0040, 128, 0, False)

    def test_overlap_subsume_after(self):
        # Existing segments:   |_|_|_|_|
        # New segment:       |_______|
        prog = Program(MOCK_PLATFORM)
        segment1 = unittest.mock.Mock(side_effect=zero_memory_read)
        segment2 = unittest.mock.Mock(side_effect=zero_memory_read)
        segment3 = unittest.mock.Mock(side_effect=zero_memory_read)
        prog.add_memory_segment(0xFFFF0020, 32, segment1)
        prog.add_memory_segment(0xFFFF0040, 32, segment1)
        prog.add_memory_segment(0xFFFF0060, 32, segment1)
        prog.add_memory_segment(0xFFFF0080, 64, segment2)
        prog.add_memory_segment(0xFFFF0000, 128, segment3)
        prog.read(0xFFFF0000, 192)
        segment1.assert_not_called()
        segment2.assert_called_once_with(0xFFFF0080, 64, 0, False)
        segment3.assert_called_once_with(0xFFFF0000, 128, 0, False)

    def test_overlap_segment_head(self):
        # Existing segment:     |_______|
        # New segment:      |_______|
        prog = Program(MOCK_PLATFORM)
        segment1 = unittest.mock.Mock(side_effect=zero_memory_read)
        segment2 = unittest.mock.Mock(side_effect=zero_memory_read)
        prog.add_memory_segment(0xFFFF0040, 128, segment1)
        prog.add_memory_segment(0xFFFF0000, 128, segment2)
        prog.read(0xFFFF0000, 192)
        segment1.assert_called_once_with(0xFFFF0080, 64, 64, False)
        segment2.assert_called_once_with(0xFFFF0000, 128, 0, False)

    def test_overlap_segment_head_and_tail(self):
        # Existing segment: |_______||_______|
        # New segment:          |_______|
        prog = Program(MOCK_PLATFORM)
        segment1 = unittest.mock.Mock(side_effect=zero_memory_read)
        segment2 = unittest.mock.Mock(side_effect=zero_memory_read)
        segment3 = unittest.mock.Mock(side_effect=zero_memory_read)
        prog.add_memory_segment(0xFFFF0000, 128, segment1)
        prog.add_memory_segment(0xFFFF0080, 128, segment2)
        prog.add_memory_segment(0xFFFF0040, 128, segment3)
        prog.read(0xFFFF0000, 256)
        segment1.assert_called_once_with(0xFFFF0000, 64, 0, False)
        segment2.assert_called_once_with(0xFFFF00C0, 64, 64, False)
        segment3.assert_called_once_with(0xFFFF0040, 128, 0, False)

    def test_overlap_subsume_at_and_after(self):
        # Existing segments: |_|_|_|_|
        # New segment:       |_______|
        prog = Program(MOCK_PLATFORM)
        segment1 = unittest.mock.Mock(side_effect=zero_memory_read)
        segment2 = unittest.mock.Mock(side_effect=zero_memory_read)
        prog.add_memory_segment(0xFFFF0000, 32, segment1)
        prog.add_memory_segment(0xFFFF0020, 32, segment1)
        prog.add_memory_segment(0xFFFF0040, 32, segment1)
        prog.add_memory_segment(0xFFFF0060, 32, segment1)
        prog.add_memory_segment(0xFFFF0000, 128, segment2)
        prog.read(0xFFFF0000, 128)
        segment1.assert_not_called()
        segment2.assert_called_once_with(0xFFFF0000, 128, 0, False)

    def test_invalid_read_fn(self):
        prog = mock_program()

        self.assertRaises(TypeError, prog.add_memory_segment, 0xFFFF0000, 8, b"foo")

        prog.add_memory_segment(0xFFFF0000, 8, lambda: None)
        self.assertRaises(TypeError, prog.read, 0xFFFF0000, 8)

        prog.add_memory_segment(
            0xFFFF0000, 8, lambda address, count, offset, physical: None
        )
        self.assertRaises(TypeError, prog.read, 0xFFFF0000, 8)

        prog.add_memory_segment(
            0xFFFF0000, 8, lambda address, count, offset, physical: "asdf"
        )
        self.assertRaises(TypeError, prog.read, 0xFFFF0000, 8)

        prog.add_memory_segment(
            0xFFFF0000, 8, lambda address, count, offset, physical: b""
        )
        self.assertRaisesRegex(
            ValueError,
            r"memory read callback returned buffer of length 0 \(expected 8\)",
            prog.read,
            0xFFFF0000,
            8,
        )

    def test_python_fault_error(self):
        def fault_memory_reader(address, count, offset, physical):
            raise FaultError("fault from Python", address)

        prog = Program(MOCK_PLATFORM)
        prog.add_memory_segment(0xFFFF0000, 8, fault_memory_reader)

        with self.assertRaises(FaultError) as cm:
            Object(prog, "int", address=0xFFFF0004).read_()
        self.assertEqual(cm.exception.message, "fault from Python")
        self.assertEqual(cm.exception.address, 0xFFFF0004)

        # If the FaultError from Python is translated to a drgn_error
        # correctly, then this shouldn't raise an exception.
        str(Object(prog, "int *", 0xFFFF0004))

    def test_python_fault_error_invalid_message(self):
        def fault_memory_reader(address, count, offset, physical):
            raise FaultError(None, address)

        prog = Program(MOCK_PLATFORM)
        prog.add_memory_segment(0xFFFF0000, 8, fault_memory_reader)

        # Just test that it doesn't crash.
        self.assertRaises(Exception, Object(prog, "int", address=0xFFFF0004).read_)

    def test_python_fault_error_invalid_address(self):
        def fault_memory_reader(address, count, offset, physical):
            raise FaultError("fault from Python", None)

        prog = Program(MOCK_PLATFORM)
        prog.add_memory_segment(0xFFFF0000, 8, fault_memory_reader)

        # Just test that it doesn't crash.
        self.assertRaises(Exception, Object(prog, "int", address=0xFFFF0004).read_)


class TestTypeFinder(TestCase):
    def test_register(self):
        prog = Program(MOCK_PLATFORM)

        # We don't test every corner case because the symbol finder tests cover
        # the shared part.
        self.assertEqual(prog.registered_type_finders(), {"dwarf"})
        self.assertEqual(prog.enabled_type_finders(), ["dwarf"])

        prog.register_type_finder(
            "foo", lambda prog, kinds, name, filename: None, enable_index=-1
        )
        self.assertEqual(prog.registered_type_finders(), {"dwarf", "foo"})
        self.assertEqual(prog.enabled_type_finders(), ["dwarf", "foo"])

        prog.set_enabled_type_finders(["foo"])
        self.assertEqual(prog.registered_type_finders(), {"dwarf", "foo"})
        self.assertEqual(prog.enabled_type_finders(), ["foo"])

    def test_add_type_finder(self):
        prog = Program(MOCK_PLATFORM)

        def dummy(kind, name, filename):
            if kind == TypeKind.TYPEDEF and name == "foo":
                return prog.typedef_type("foo", prog.void_type())
            else:
                return None

        prog.add_type_finder(dummy)
        self.assertTrue(any("dummy" in name for name in prog.registered_type_finders()))
        self.assertIn("dummy", prog.enabled_type_finders()[0])
        self.assertIdentical(
            prog.type("foo"), prog.typedef_type("foo", prog.void_type())
        )

    def test_register_invalid(self):
        prog = Program(MOCK_PLATFORM)
        self.assertRaises(TypeError, prog.register_type_finder, "foo", "foo")
        prog.register_type_finder(
            "foo", lambda prog, kinds, name, filename: "foo", enable_index=0
        )
        self.assertRaises(TypeError, prog.type, "int")

    def test_add_invalid(self):
        prog = Program(MOCK_PLATFORM)
        self.assertRaises(TypeError, prog.add_type_finder, "foo")
        prog.add_type_finder(lambda kind, name, filename: "foo")
        self.assertRaises(TypeError, prog.type, "int")

    def test_register_wrong_program(self):
        def finder(prog, kinds, name, filename):
            if TypeKind.TYPEDEF in kinds and name == "foo":
                prog = Program()
                return prog.typedef_type("foo", prog.void_type())
            else:
                return None

        prog = Program(MOCK_PLATFORM)
        prog.register_type_finder("foo", finder, enable_index=0)
        self.assertRaisesRegex(
            ValueError,
            "type find callback returned type from wrong program",
            prog.type,
            "foo",
        )

    def test_add_wrong_program(self):
        def finder(kind, name, filename):
            if kind == TypeKind.TYPEDEF and name == "foo":
                prog = Program()
                return prog.typedef_type("foo", prog.void_type())
            else:
                return None

        prog = Program(MOCK_PLATFORM)
        prog.add_type_finder(finder)
        self.assertRaisesRegex(
            ValueError,
            "type find callback returned type from wrong program",
            prog.type,
            "foo",
        )

    def test_register_wrong_kind(self):
        prog = Program(MOCK_PLATFORM)
        prog.register_type_finder(
            "foo", lambda prog, kinds, name, filename: prog.void_type(), enable_index=0
        )
        self.assertRaises(TypeError, prog.type, "int")

    def test_add_wrong_kind(self):
        prog = Program(MOCK_PLATFORM)
        prog.add_type_finder(lambda kind, name, filename: prog.void_type())
        self.assertRaises(TypeError, prog.type, "int")

    def test_register_not_found(self):
        prog = Program(MOCK_PLATFORM)
        prog.register_type_finder(
            "foo", lambda prog, kinds, name, filename: None, enable_index=0
        )
        self.assertRaises(LookupError, prog.type, "struct foo")

    def test_add_not_found(self):
        prog = Program(MOCK_PLATFORM)
        prog.add_type_finder(lambda kind, name, filename: None)
        self.assertRaises(LookupError, prog.type, "struct foo")


class TestObjectFinder(TestCase):
    def test_register(self):
        prog = Program(MOCK_PLATFORM)

        # We don't test every corner case because the symbol finder tests cover
        # the shared part.
        self.assertEqual(prog.registered_object_finders(), {"dwarf"})
        self.assertEqual(prog.enabled_object_finders(), ["dwarf"])

        prog.register_object_finder(
            "foo", lambda prog, name, flags, filename: None, enable_index=-1
        )
        self.assertEqual(prog.registered_object_finders(), {"dwarf", "foo"})
        self.assertEqual(prog.enabled_object_finders(), ["dwarf", "foo"])

        prog.set_enabled_object_finders(["foo"])
        self.assertEqual(prog.registered_object_finders(), {"dwarf", "foo"})
        self.assertEqual(prog.enabled_object_finders(), ["foo"])

    def test_add_object_finder(self):
        prog = Program(MOCK_PLATFORM)

        def dummy(prog, name, flags, filename):
            return Object(prog, "int", 1)

        prog.add_object_finder(dummy)
        self.assertTrue(
            any("dummy" in name for name in prog.registered_object_finders())
        )
        self.assertIn("dummy", prog.enabled_object_finders()[0])
        self.assertIdentical(prog.object("foo"), Object(prog, "int", 1))

    def test_register_invalid(self):
        prog = Program(MOCK_PLATFORM)
        self.assertRaises(TypeError, prog.register_object_finder, "foo", "foo")
        prog.register_object_finder(
            "foo", lambda prog, name, flags, filename: "foo", enable_index=0
        )
        self.assertRaises(TypeError, prog.object, "foo")

    def test_add_invalid(self):
        prog = Program(MOCK_PLATFORM)
        self.assertRaises(TypeError, prog.add_object_finder, "foo")
        prog.add_object_finder(lambda prog, name, flags, filename: "foo")
        self.assertRaises(TypeError, prog.object, "foo")

    def test_wrong_program(self):
        prog = Program(MOCK_PLATFORM)
        prog.register_object_finder(
            "foo",
            lambda prog, name, flags, filename: Object(
                Program(MOCK_PLATFORM), "int", 1
            ),
            enable_index=0,
        )
        self.assertRaisesRegex(
            ValueError,
            "different program",
            prog.object,
            "foo",
        )

    def test_not_found(self):
        prog = Program(MOCK_PLATFORM)
        self.assertRaises(LookupError, prog.object, "foo")
        prog.register_object_finder(
            "foo", lambda prog, name, flags, filename: None, enable_index=0
        )
        self.assertRaises(LookupError, prog.object, "foo")
        self.assertFalse("foo" in prog)


class TestTypes(MockProgramTestCase):
    def test_already_type(self):
        self.assertIdentical(
            self.prog.type(self.prog.pointer_type(self.prog.void_type())),
            self.prog.pointer_type(self.prog.void_type()),
        )

    def test_invalid_argument_type(self):
        self.assertRaises(TypeError, self.prog.type, 1)

    def test_default_primitive_types(self):
        def spellings(tokens, num_optional=0):
            for i in range(len(tokens) - num_optional, len(tokens) + 1):
                for perm in itertools.permutations(tokens[:i]):
                    yield " ".join(perm)

        for word_size in [8, 4]:
            prog = mock_program(
                MOCK_PLATFORM if word_size == 8 else MOCK_32BIT_PLATFORM
            )
            self.assertIdentical(prog.type("_Bool"), prog.bool_type("_Bool", 1))
            self.assertIdentical(prog.type("char"), prog.int_type("char", 1, True))
            for spelling in spellings(["signed", "char"]):
                self.assertIdentical(
                    prog.type(spelling), prog.int_type("signed char", 1, True)
                )
            for spelling in spellings(["unsigned", "char"]):
                self.assertIdentical(
                    prog.type(spelling), prog.int_type("unsigned char", 1, False)
                )
            for spelling in spellings(["short", "signed", "int"], 2):
                self.assertIdentical(
                    prog.type(spelling), prog.int_type("short", 2, True)
                )
            for spelling in spellings(["short", "unsigned", "int"], 1):
                self.assertIdentical(
                    prog.type(spelling), prog.int_type("unsigned short", 2, False)
                )
            for spelling in spellings(["int", "signed"], 1):
                self.assertIdentical(prog.type(spelling), prog.int_type("int", 4, True))
            for spelling in spellings(["unsigned", "int"]):
                self.assertIdentical(
                    prog.type(spelling), prog.int_type("unsigned int", 4, False)
                )
            for spelling in spellings(["long", "signed", "int"], 2):
                self.assertIdentical(
                    prog.type(spelling), prog.int_type("long", word_size, True)
                )
            for spelling in spellings(["long", "unsigned", "int"], 1):
                self.assertIdentical(
                    prog.type(spelling),
                    prog.int_type("unsigned long", word_size, False),
                )
            for spelling in spellings(["long", "long", "signed", "int"], 2):
                self.assertIdentical(
                    prog.type(spelling), prog.int_type("long long", 8, True)
                )
            for spelling in spellings(["long", "long", "unsigned", "int"], 1):
                self.assertIdentical(
                    prog.type(spelling), prog.int_type("unsigned long long", 8, False)
                )
            self.assertIdentical(prog.type("float"), prog.float_type("float", 4))
            self.assertIdentical(prog.type("double"), prog.float_type("double", 8))
            for spelling in spellings(["long", "double"]):
                self.assertIdentical(
                    prog.type(spelling), prog.float_type("long double", 16)
                )
            self.assertIdentical(
                prog.type("size_t"),
                prog.typedef_type(
                    "size_t", prog.int_type("unsigned long", word_size, False)
                ),
            )
            self.assertIdentical(
                prog.type("ptrdiff_t"),
                prog.typedef_type("ptrdiff_t", prog.int_type("long", word_size, True)),
            )

    def test_primitive_type(self):
        self.types.append(self.prog.int_type("long", 4, True))
        self.assertIdentical(
            self.prog.type("long"), self.prog.int_type("long", 4, True)
        )

    def test_primitive_type_invalid(self):
        # unsigned long with signed=True isn't valid, so it should be ignored.
        self.types.append(self.prog.int_type("unsigned long", 4, True))
        self.assertIdentical(
            self.prog.type("unsigned long"),
            self.prog.int_type("unsigned long", 8, False),
        )

    def test_size_t_and_ptrdiff_t(self):
        # 64-bit architecture with 4-byte long/unsigned long.
        types = []
        prog = mock_program(types=types)
        types.append(prog.int_type("long", 4, True))
        types.append(prog.int_type("unsigned long", 4, False))
        self.assertIdentical(
            prog.type("size_t"),
            prog.typedef_type("size_t", prog.type("unsigned long long")),
        )
        self.assertIdentical(
            prog.type("ptrdiff_t"),
            prog.typedef_type("ptrdiff_t", prog.type("long long")),
        )

        # 32-bit architecture with 8-byte long/unsigned long.
        types = []
        prog = mock_program(MOCK_32BIT_PLATFORM, types=types)
        types.append(prog.int_type("long", 8, True))
        types.append(prog.int_type("unsigned long", 8, False))
        self.assertIdentical(
            prog.type("size_t"), prog.typedef_type("size_t", prog.type("unsigned int"))
        )
        self.assertIdentical(
            prog.type("ptrdiff_t"), prog.typedef_type("ptrdiff_t", prog.type("int"))
        )

        # Nonsense sizes.
        types = []
        prog = mock_program(types=types)
        types.append(prog.int_type("int", 1, True))
        types.append(prog.int_type("unsigned int", 1, False))
        types.append(prog.int_type("long", 1, True))
        types.append(prog.int_type("unsigned long", 1, False))
        types.append(prog.int_type("long long", 2, True))
        types.append(prog.int_type("unsigned long long", 2, False))
        self.assertRaisesRegex(
            ValueError, "no suitable integer type for size_t", prog.type, "size_t"
        )
        self.assertRaisesRegex(
            ValueError, "no suitable integer type for ptrdiff_t", prog.type, "ptrdiff_t"
        )

    def test_not_size_t_or_ptrdiff_t(self):
        self.types.append(
            self.prog.typedef_type(
                "size_tea", self.prog.int_type("unsigned char", 1, False)
            )
        )
        self.types.append(
            self.prog.typedef_type("ptrdiff_tee", self.prog.int_type("char", 1, True))
        )
        self.assertIdentical(
            self.prog.type("size_tea"),
            self.prog.typedef_type(
                "size_tea", self.prog.int_type("unsigned char", 1, False)
            ),
        )
        self.assertIdentical(
            self.prog.type("ptrdiff_tee"),
            self.prog.typedef_type("ptrdiff_tee", self.prog.int_type("char", 1, True)),
        )

    def test_tagged_type(self):
        self.types.append(self.point_type)
        self.types.append(self.option_type)
        self.types.append(self.color_type)
        self.assertIdentical(self.prog.type("struct point"), self.point_type)
        self.assertIdentical(self.prog.type("union option"), self.option_type)
        self.assertIdentical(self.prog.type("enum color"), self.color_type)

    def test_class_type(self):
        struct_class = self.prog.struct_type(
            "class",
            8,
            (TypeMember(self.prog.pointer_type(self.prog.void_type()), "ptr"),),
        )
        class_point = self.prog.class_type(
            "Point",
            8,
            (
                TypeMember(self.prog.int_type("int", 4, True), "x", 0),
                TypeMember(self.prog.int_type("int", 4, True), "y", 32),
            ),
        )
        self.types.append(struct_class)
        self.types.append(class_point)
        self.prog.language = Language.C
        self.assertIdentical(self.prog.type("struct class"), struct_class)
        self.prog.language = Language.CPP
        self.assertRaisesRegex(
            SyntaxError,
            "expected identifier after 'struct'",
            self.prog.type,
            "struct class",
        )
        self.assertIdentical(self.prog.type("class Point"), class_point)

    def test_typedef(self):
        self.types.append(self.pid_type)
        self.assertIdentical(self.prog.type("pid_t"), self.pid_type)

    def test_pointer(self):
        self.assertIdentical(
            self.prog.type("int *"),
            self.prog.pointer_type(self.prog.int_type("int", 4, True)),
        )

    def test_pointer_to_const(self):
        self.assertIdentical(
            self.prog.type("const int *"),
            self.prog.pointer_type(
                self.prog.int_type("int", 4, True, qualifiers=Qualifiers.CONST)
            ),
        )

    def test_const_pointer(self):
        self.assertIdentical(
            self.prog.type("int * const"),
            self.prog.pointer_type(
                self.prog.int_type("int", 4, True), qualifiers=Qualifiers.CONST
            ),
        )

    def test_pointer_to_pointer(self):
        self.assertIdentical(
            self.prog.type("int **"),
            self.prog.pointer_type(
                self.prog.pointer_type(self.prog.int_type("int", 4, True))
            ),
        )
        self.assertIdentical(self.prog.type("int *((*))"), self.prog.type("int **"))

    def test_pointer_to_const_pointer(self):
        self.assertIdentical(
            self.prog.type("int * const *"),
            self.prog.pointer_type(
                self.prog.pointer_type(
                    self.prog.int_type("int", 4, True), qualifiers=Qualifiers.CONST
                )
            ),
        )

    def test_array(self):
        self.assertIdentical(
            self.prog.type("int [20]"),
            self.prog.array_type(self.prog.int_type("int", 4, True), 20),
        )

    def test_array_hexadecimal(self):
        self.assertIdentical(
            self.prog.type("int [0x20]"),
            self.prog.array_type(self.prog.int_type("int", 4, True), 32),
        )

    def test_array_octal(self):
        self.assertIdentical(
            self.prog.type("int [020]"),
            self.prog.array_type(self.prog.int_type("int", 4, True), 16),
        )

    def test_incomplete_array(self):
        self.assertIdentical(
            self.prog.type("int []"),
            self.prog.array_type(self.prog.int_type("int", 4, True)),
        )

    def test_array_two_dimensional(self):
        self.assertIdentical(
            self.prog.type("int [2][3]"),
            self.prog.array_type(
                self.prog.array_type(self.prog.int_type("int", 4, True), 3), 2
            ),
        )

    def test_array_three_dimensional(self):
        self.assertIdentical(
            self.prog.type("int [2][3][4]"),
            self.prog.array_type(
                self.prog.array_type(
                    self.prog.array_type(self.prog.int_type("int", 4, True), 4), 3
                ),
                2,
            ),
        )

    def test_array_of_pointers(self):
        self.assertIdentical(
            self.prog.type("int *[2][3]"),
            self.prog.array_type(
                self.prog.array_type(
                    self.prog.pointer_type(self.prog.int_type("int", 4, True)), 3
                ),
                2,
            ),
        )

    def test_pointer_to_array(self):
        self.assertIdentical(
            self.prog.type("int (*)[2]"),
            self.prog.pointer_type(
                self.prog.array_type(self.prog.int_type("int", 4, True), 2)
            ),
        )

    def test_pointer_to_two_dimensional_array(self):
        self.assertIdentical(
            self.prog.type("int (*)[2][3]"),
            self.prog.pointer_type(
                self.prog.array_type(
                    self.prog.array_type(self.prog.int_type("int", 4, True), 3), 2
                )
            ),
        )

    def test_pointer_to_pointer_to_array(self):
        self.assertIdentical(
            self.prog.type("int (**)[2]"),
            self.prog.pointer_type(
                self.prog.pointer_type(
                    self.prog.array_type(self.prog.int_type("int", 4, True), 2)
                )
            ),
        )

    def test_pointer_to_array_of_pointers(self):
        self.assertIdentical(
            self.prog.type("int *(*)[2]"),
            self.prog.pointer_type(
                self.prog.array_type(
                    self.prog.pointer_type(self.prog.int_type("int", 4, True)), 2
                )
            ),
        )
        self.assertIdentical(
            self.prog.type("int *((*)[2])"), self.prog.type("int *(*)[2]")
        )

    def test_array_of_pointers_to_array(self):
        self.assertIdentical(
            self.prog.type("int (*[2])[3]"),
            self.prog.array_type(
                self.prog.pointer_type(
                    self.prog.array_type(self.prog.int_type("int", 4, True), 3)
                ),
                2,
            ),
        )


class TestObjects(MockProgramTestCase):
    def test_constant(self):
        self.objects.append(
            MockObject("PAGE_SIZE", self.prog.int_type("int", 4, True), value=4096)
        )
        self.assertIdentical(
            self.prog["PAGE_SIZE"],
            Object(self.prog, self.prog.int_type("int", 4, True), value=4096),
        )
        self.assertIdentical(
            self.prog.object("PAGE_SIZE", FindObjectFlags.CONSTANT),
            self.prog["PAGE_SIZE"],
        )
        self.assertTrue("PAGE_SIZE" in self.prog)

    def test_function(self):
        self.objects.append(
            MockObject(
                "func",
                self.prog.function_type(self.prog.void_type(), (), False),
                address=0xFFFF0000,
            )
        )
        self.assertIdentical(
            self.prog["func"],
            Object(
                self.prog,
                self.prog.function_type(self.prog.void_type(), (), False),
                address=0xFFFF0000,
            ),
        )
        self.assertIdentical(
            self.prog.object("func", FindObjectFlags.FUNCTION), self.prog["func"]
        )
        self.assertTrue("func" in self.prog)

    def test_variable(self):
        self.objects.append(
            MockObject(
                "counter", self.prog.int_type("int", 4, True), address=0xFFFF0000
            )
        )
        self.assertIdentical(
            self.prog["counter"],
            Object(self.prog, self.prog.int_type("int", 4, True), address=0xFFFF0000),
        )
        self.assertIdentical(
            self.prog.object("counter", FindObjectFlags.VARIABLE), self.prog["counter"]
        )
        self.assertTrue("counter" in self.prog)


class TestCoreDump(TestCase):
    def test_not_core_dump(self):
        prog = Program()
        self.assertRaisesRegex(
            ValueError, "not an ELF core file", prog.set_core_dump, "/dev/null"
        )
        with tempfile.NamedTemporaryFile() as f:
            f.write(create_elf_file(ET.EXEC, []))
            f.flush()
            self.assertRaisesRegex(
                ValueError, "not an ELF core file", prog.set_core_dump, f.name
            )

    def test_twice(self):
        prog = Program()
        with tempfile.NamedTemporaryFile() as f:
            f.write(create_elf_file(ET.CORE, []))
            f.flush()
            prog.set_core_dump(f.name)
            self.assertRaisesRegex(
                ValueError,
                "program memory was already initialized",
                prog.set_core_dump,
                f.name,
            )

    def test_simple(self):
        data = b"hello, world"
        prog = Program()
        self.assertIsNone(prog.core_dump_path)
        with tempfile.NamedTemporaryFile() as f:
            f.write(
                create_elf_file(
                    ET.CORE, [ElfSection(p_type=PT.LOAD, vaddr=0xFFFF0000, data=data)]
                )
            )
            f.flush()
            prog.set_core_dump(f.name)
        self.assertEqual(prog.core_dump_path, f.name)
        self.assertEqual(prog.read(0xFFFF0000, len(data)), data)
        self.assertRaises(FaultError, prog.read, 0x0, len(data), physical=True)

    def test_physical(self):
        data = b"hello, world"
        prog = Program()
        with tempfile.NamedTemporaryFile() as f:
            f.write(
                create_elf_file(
                    ET.CORE,
                    [
                        ElfSection(
                            p_type=PT.LOAD, vaddr=0xFFFF0000, paddr=0xA0, data=data
                        ),
                    ],
                )
            )
            f.flush()
            prog.set_core_dump(f.name)
        self.assertEqual(prog.read(0xFFFF0000, len(data)), data)
        self.assertEqual(prog.read(0xA0, len(data), physical=True), data)

    def test_unsaved(self):
        data = b"hello, world"
        prog = Program()
        with tempfile.NamedTemporaryFile() as f:
            f.write(
                create_elf_file(
                    ET.CORE,
                    [
                        ElfSection(
                            p_type=PT.LOAD,
                            vaddr=0xFFFF0000,
                            data=data,
                            memsz=len(data) + 4,
                        ),
                    ],
                )
            )
            f.flush()
            prog.set_core_dump(f.name)
        with self.assertRaisesRegex(FaultError, "memory not saved in core dump") as cm:
            prog.read(0xFFFF0000, len(data) + 4)
        self.assertEqual(cm.exception.address, 0xFFFF000C)


def make_vmcoreinfo(
    osrelease="6.0.0-test",
    pagesize=4096,
    swapper_pg_dir=0xFFFFFFFF81000000,
    extra_lines=None,
):
    """Create fake vmcoreinfo data with required fields."""
    lines = [
        f"OSRELEASE={osrelease}",
        f"PAGESIZE={pagesize}",
        f"SYMBOL(swapper_pg_dir)={swapper_pg_dir:x}",
    ]
    if extra_lines:
        lines.extend(extra_lines)
    return ("\n".join(lines) + "\n").encode()


class TestSetLinuxKernelCustom(TestCase):
    def test_requires_platform(self):
        prog = Program()
        self.assertRaisesRegex(
            ValueError,
            "platform must be set",
            prog.set_linux_kernel_custom,
            make_vmcoreinfo(),
            False,
        )

    def test_invalid_vmcoreinfo_missing_osrelease(self):
        prog = Program(MOCK_PLATFORM)
        self.assertRaisesRegex(
            Exception,
            "VMCOREINFO does not contain valid OSRELEASE",
            prog.set_linux_kernel_custom,
            b"PAGESIZE=4096\nSYMBOL(swapper_pg_dir)=ffff0000\n",
            False,
        )

    def test_sets_linux_kernel_flag(self):
        prog = Program(MOCK_PLATFORM)
        self.assertFalse(prog.flags & ProgramFlags.IS_LINUX_KERNEL)
        prog.set_linux_kernel_custom(make_vmcoreinfo(), False)
        self.assertTrue(prog.flags & ProgramFlags.IS_LINUX_KERNEL)

    def test_is_live(self):
        prog = Program(MOCK_PLATFORM)
        self.assertFalse(prog.flags & ProgramFlags.IS_LIVE)
        prog.set_linux_kernel_custom(make_vmcoreinfo(), True)
        self.assertTrue(prog.flags & ProgramFlags.IS_LIVE)

    def test_not_is_live(self):
        prog = Program(MOCK_PLATFORM)
        prog.set_linux_kernel_custom(make_vmcoreinfo(), False)
        self.assertFalse(prog.flags & ProgramFlags.IS_LIVE)

    def test_idempotent(self):
        prog = Program(MOCK_PLATFORM)
        prog.set_linux_kernel_custom(make_vmcoreinfo(), False)
        self.assertTrue(prog.flags & ProgramFlags.IS_LINUX_KERNEL)
        prog.set_linux_kernel_custom(make_vmcoreinfo(), False)
        self.assertTrue(prog.flags & ProgramFlags.IS_LINUX_KERNEL)

    def test_with_vmcoreinfo_in_constructor(self):
        vmcoreinfo1 = make_vmcoreinfo(osrelease="5.0.0-first")
        vmcoreinfo2 = make_vmcoreinfo(osrelease="6.0.0-second")
        prog = Program(MOCK_PLATFORM, vmcoreinfo=vmcoreinfo1)
        prog.set_linux_kernel_custom(vmcoreinfo2, False)
        self.assertTrue(prog.flags & ProgramFlags.IS_LINUX_KERNEL)

    def test_with_physical_memory_segment(self):
        prog = Program(MOCK_PLATFORM)
        data = b"test data"
        prog.add_memory_segment(
            0x1000,
            len(data),
            lambda addr, count, off, phys: data[off : off + count],
            True,
        )
        prog.set_linux_kernel_custom(make_vmcoreinfo(), False)
        self.assertTrue(prog.flags & ProgramFlags.IS_LINUX_KERNEL)
        self.assertEqual(prog.read(0x1000, len(data), physical=True), data)


def dummy_symbol_finder(prog, name, address, one):
    return ()


class TestSymbolFinders(TestCase):
    def test_registered(self):
        prog = Program()
        self.assertEqual(prog.registered_symbol_finders(), {"elf"})
        prog.register_symbol_finder("foo", dummy_symbol_finder)
        self.assertEqual(prog.registered_symbol_finders(), {"elf", "foo"})

    def test_register_duplicate(self):
        self.assertRaisesRegex(
            ValueError,
            "duplicate symbol finder",
            Program().register_symbol_finder,
            "elf",
            dummy_symbol_finder,
        )

    def test_default_enabled(self):
        self.assertEqual(Program().enabled_symbol_finders(), ["elf"])

    def test_disable_all(self):
        prog = Program()
        prog.set_enabled_symbol_finders(())
        self.assertEqual(prog.enabled_symbol_finders(), [])

    def test_register_then_enable(self):
        prog = Program()
        prog.register_symbol_finder("foo", dummy_symbol_finder)
        self.assertEqual(prog.enabled_symbol_finders(), ["elf"])

        prog.set_enabled_symbol_finders(["foo", "elf"])
        self.assertEqual(prog.enabled_symbol_finders(), ["foo", "elf"])

    def test_register_enable_index(self):
        prog = Program()
        with self.subTest("None"):
            prog.register_symbol_finder("ghost", dummy_symbol_finder, enable_index=None)
            self.assertEqual(prog.enabled_symbol_finders(), ["elf"])

        with self.subTest("first"):
            prog.register_symbol_finder("foo", dummy_symbol_finder, enable_index=0)
            self.assertEqual(prog.enabled_symbol_finders(), ["foo", "elf"])

        with self.subTest("middle"):
            prog.register_symbol_finder("bar", dummy_symbol_finder, enable_index=1)
            self.assertEqual(prog.enabled_symbol_finders(), ["foo", "bar", "elf"])

        with self.subTest("end"):
            prog.register_symbol_finder("baz", dummy_symbol_finder, enable_index=3)
            self.assertEqual(
                prog.enabled_symbol_finders(), ["foo", "bar", "elf", "baz"]
            )

        with self.subTest("past end"):
            prog.register_symbol_finder("qux", dummy_symbol_finder, enable_index=10)
            self.assertEqual(
                prog.enabled_symbol_finders(), ["foo", "bar", "elf", "baz", "qux"]
            )

        with self.subTest("DRGN_HANDLER_REGISTER_DONT_ENABLE"):
            prog.register_symbol_finder(
                "quux",
                dummy_symbol_finder,
                enable_index=(2**64 if sys.maxsize > 2**32 else 2**32) - 2,
            )
            self.assertEqual(
                prog.enabled_symbol_finders(),
                ["foo", "bar", "elf", "baz", "qux", "quux"],
            )

        with self.subTest("last"):
            prog.register_symbol_finder("quuux", dummy_symbol_finder, enable_index=-1)
            self.assertEqual(
                prog.enabled_symbol_finders(),
                ["foo", "bar", "elf", "baz", "qux", "quux", "quuux"],
            )

    def test_register_enable_index_invalid(self):
        self.assertRaises(
            OverflowError,
            Program().register_symbol_finder,
            "foo",
            dummy_symbol_finder,
            enable_index=-2,
        )
